"""Functions for the colocalization analysis in pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
from functools import partial
from typing import Optional, get_args

import pandas as pd
from statsmodels.stats.multitest import multipletests

from pixelator.common.statistics import (
    correct_pvalues,
    log1p_transformation,
    rate_diff_transformation,
    wilcoxon_test,
)
from pixelator.mpx.analysis.analysis_engine import PerComponentAnalysis
from pixelator.mpx.analysis.colocalization.estimate import (
    estimate_observation_statistics,
    permutation_analysis_results,
)
from pixelator.mpx.analysis.colocalization.prepare import (
    filter_by_region_counts,
    filter_by_unique_values,
    prepare_from_graph,
)
from pixelator.mpx.analysis.colocalization.statistics import (
    Jaccard,
    Pearson,
    apply_multiple_stats,
)
from pixelator.mpx.analysis.colocalization.types import (
    MarkerColocalizationResults,
    TransformationTypes,
)
from pixelator.mpx.analysis.permute import permutations
from pixelator.mpx.graph.utils import Graph
from pixelator.mpx.pixeldataset import PixelDataset

logger = logging.getLogger(__name__)


def colocalization_from_component_edgelist(
    edgelist: pd.DataFrame,
    component_id: str,
    transformation: TransformationTypes = "raw",
    neighbourhood_size: int = 0,
    n_permutations: int = 50,
    use_full_bipartite: bool = True,
    min_region_count: int = 5,
    min_marker_count: int = 5,
    random_seed: Optional[int] = None,
) -> pd.DataFrame:
    """Get the colocalization scores for the component in the given `edgelist`.

    :param edgelist: edgelist to compute colocalization scores for
    :param component_id: name of the component
    :param transformation: transformation method to use, defaults to "raw"
    :param neighbourhood_size: size of the neighbourhood to consider, defaults to 0
    :param n_permutations: number of permutations used to calculate the
                           p-values and z-scores, defaults to 50
    :param use_full_bipartite: use the full bipartiate graph, if false use A-node
                               projection, defaults to True
    :param min_region_count: minimum number of counts in region to consider, defaults
                             to 5
    :param min_marker_count: the minimum number of counts of a marker to calculate
                             colocalization
    :param random_seed: Set the random seed for the permutation tests, defaults to None
    :return: a dataframe with computed colocalization scores
    :rtype: pd.DataFrame
    """
    graph = Graph.from_edgelist(
        edgelist=edgelist,
        add_marker_counts=True,
        # If we do A-node projection, we will simplify anyway.
        # This just removes the warning.
        simplify=False if use_full_bipartite else True,
        use_full_bipartite=use_full_bipartite,
    )
    return colocalization_from_component_graph(
        graph=graph,
        component_id=component_id,
        transformation=transformation,
        neighbourhood_size=neighbourhood_size,
        n_permutations=n_permutations,
        min_region_count=min_region_count,
        min_marker_count=min_marker_count,
        random_seed=random_seed,
    )


def _transform_data(
    data: MarkerColocalizationResults, transform: TransformationTypes
) -> MarkerColocalizationResults:
    if transform == "raw":
        return data
    if transform == "log1p":
        return log1p_transformation(data)
    if transform == "rate-diff":
        return rate_diff_transformation(data)
    raise ValueError(
        f"`transform`must be one of: {'/'.join(get_args(TransformationTypes))}"
    )


def colocalization_from_component_graph(
    graph: Graph,
    component_id: str,
    transformation: TransformationTypes = "raw",
    neighbourhood_size: int = 1,
    n_permutations: int = 50,
    min_region_count: int = 5,
    min_marker_count: int = 5,
    random_seed: Optional[int] = None,
) -> pd.DataFrame:
    """Compute the colocalization scores for this component graph.

    :param graph: graph to compute scores for
    :param component_id: name of the component
    :param transformation: transformation method to use, defaults to "raw"
    :param neighbourhood_size: size of the neighbourhood to consider, defaults to 0
    :param n_permutations: number of permutations used to calculate the
                           p-values and z-scores, defaults to 50
    :param min_region_count: minimum number of counts in region to consider, defaults
                             to 5
    :param min_marker_count: the minimum number of counts of a marker to calculate
                             colocalization
    :param random_seed: Set the random seed for the permutation tests, defaults to None
    :return: a dataframe containing colocalization scores for this component
    :rtype: pd.DataFrame
    """
    logger.debug("Computing colocalization for component: %s", component_id)
    logger.debug("Prepare the graph data for computing colocalization")

    raw_marker_counts = graph.node_marker_counts
    # Record markers to keep
    # Remove markers with zero variance and markers below minimum marker count
    markers_to_keep = raw_marker_counts.columns[
        (raw_marker_counts != 0).any(axis=0)
        & (raw_marker_counts.nunique() > 1)
        & (raw_marker_counts.sum() >= min_marker_count)
    ]

    marker_counts_by_region = prepare_from_graph(graph, n_neighbours=neighbourhood_size)
    marker_counts_by_region = marker_counts_by_region[markers_to_keep]
    marker_counts_by_region = filter_by_region_counts(
        marker_counts_by_region, min_region_counts=min_region_count
    )

    nrow, ncols = marker_counts_by_region.shape
    if nrow == 0:
        logger.warning(
            "Component: %s had no valid regions with the filters you set", component_id
        )
        return pd.DataFrame()
    if ncols == 0:
        logger.warning(
            "Component: %s has no valid markers with the filters you set", component_id
        )
        return pd.DataFrame()

    transformed_marker_counts_by_region = _transform_data(
        marker_counts_by_region, transformation
    )

    functions_of_interest = (Pearson, Jaccard)
    logger.debug("Computing: %s", [f.name for f in functions_of_interest])
    observations = apply_multiple_stats(
        df=transformed_marker_counts_by_region, funcs=functions_of_interest
    )

    logger.debug("Running permutation analysis")
    permutation_results = permutation_analysis_results(
        marker_counts_by_region,
        funcs=functions_of_interest,
        permuter=permutations,
        transformer=partial(_transform_data, transform=transformation),
        n=n_permutations,
        random_seed=random_seed,
    )

    logger.debug("Estimating observation statistics")
    results = estimate_observation_statistics(
        observations=observations,
        permutation_results=permutation_results,
        funcs=functions_of_interest,
    )
    results["component"] = component_id
    logger.debug("Finished computing colocalization for component: %s", component_id)

    return results


def colocalization_scores(
    edgelist: pd.DataFrame,
    use_full_bipartite: bool = True,
    transformation: TransformationTypes = "raw",
    neighbourhood_size: int = 1,
    n_permutations: int = 50,
    min_region_count: int = 5,
    min_marker_count: int = 5,
    random_seed: Optional[int] = None,
) -> pd.DataFrame:
    """Compute colocalization scores for antibody pairs.

    Computes colocalization scores (unique antibody pairs) for each component
    in the `edgelist` given as input. Only the unique combination of antibodies
    are included and the component id is present in the dataframe which has the
    following columns:
        - marker_1
        - marker_2
        - pearson
        - pearson_z
        - pearson_p_value
        - pearson_p_value_adjusted
        - jaccard
        - jaccard_z
        - jaccard_p_value
        - jaccard_p_value_adjusted
        - component
    The function iterates all the components to compute the scores for each
    pair of antibodies. The scores are computed using the Jaccard index on the
    binarize antibody counts (colocalization) or the Pearson Correlation
    Coefficient on the counts (coabundance). The scores should be high when
    two antibodies are located in the same area of the graph (taking into
    consideration size differences) and their expression is similar (Pearson).
    It also does permutation testing of each colocalizaiton measure to compute
    emprical p-values, corrected p-values, and z-score for each measure.
    :param edgelist: an edge list dataframe with a membership column
    :param use_full_bipartite: use the bipartite graph instead of the projection (UPIA)
    :param transformation: Select a transformation method to use for the colocalization
    :param neighbourhood_size: Set the size of the neighbourhood to
                               consider when computing the colocalization
    :param n_permutations: Select number of permutations used to
                           calculate empirical p-values of the
                           colocalization values
    :param min_region_count: The minimum size of the region (e.g. number
                             of counts in the neighbourhood) required
                             for it to be considered
    :param min_marker_count: the minimum number of counts of a marker to calculate
                             colocalization
    :param random_seed: Set a random seed for the permutation function
    :returns: a pd.DataFrame of scores
    :rtype: pd.DataFrame
    :raises AssertionError: when the input is not valid
    :raises ValueError: when no components were found to be valid for
                        computing colocalization.

    """
    if "component" not in edgelist.columns:
        raise AssertionError("edge list is missing the membership column")

    logger.debug(
        "Computing colocalization scores for edge list with %i elements",
        edgelist.shape[0],
    )

    def data():
        grouped = edgelist.groupby("component", observed=True)
        unique_components = len(edgelist["component"].unique())
        for idx, (component_id, component_df) in enumerate(grouped):
            logger.debug(
                "Computing colocalization of %i/%i",
                idx + 1,
                unique_components,
            )
            # build the graph from the component
            graph = Graph.from_edgelist(
                edgelist=component_df,
                add_marker_counts=True,
                # If we do A-node projection, we will simplify anyway.
                # This just removes the warning.
                simplify=False if use_full_bipartite else True,
                use_full_bipartite=use_full_bipartite,
            )
            if len(graph.vs) < 2:
                logger.warning(
                    "Component %s only had a single node. It will be skipped.",
                    component_id,
                )
                continue

            yield colocalization_from_component_graph(
                graph=graph,
                component_id=component_id,
                transformation=transformation,
                neighbourhood_size=neighbourhood_size,
                n_permutations=n_permutations,
                min_region_count=min_region_count,
                min_marker_count=min_marker_count,
                random_seed=random_seed,
            )

    # create dataframe with all the scores
    try:
        scores = pd.concat(data(), axis=0)
    except ValueError as error:
        logger.error(
            "No data was found to compute colocalization, probably "
            "because all components only had a single node."
        )
        raise error

    p_value_columns = filter(lambda x: "_p" in x, scores.columns)
    for p_value_col in p_value_columns:
        scores.insert(
            scores.columns.get_loc(p_value_col) + 1,
            f"{p_value_col}_adjusted",
            correct_pvalues(scores[p_value_col].to_numpy()),
        )

    logger.debug("Colocalization scores for dataset computed")
    return scores


def get_differential_colocalization(
    colocalization_data_frame: pd.DataFrame,
    reference: str,
    targets: str | list[str] | None = None,
    contrast_column: str = "sample",
    value_column: str = "pearson_z",
) -> pd.DataFrame:
    """Calculate the differential colocalization.

    :param colocalization_data_frame: The colocalization data frame.
    :param target: The label for target components in the contrast_column.
    :param reference: The label for reference components in the contrast_column.
    :param contrast_column: The column to use for the contrast. Defaults to "sample".
    :param value_column: What colocalization metric to use. Defaults to "pearson_z".

    :return: The differential colocalization.
    :rtype: pd.DataFrame
    """
    if targets is None:
        targets = colocalization_data_frame[contrast_column].unique()
        targets = list(set(targets) - {reference})
    elif isinstance(targets, str):
        targets = [targets]

    if len(targets) > 5:
        logger.warning(
            "There are more than 5 targets in the dataset. This may take a while."
        )
    same_marker_mask = (
        colocalization_data_frame["marker_1"] == colocalization_data_frame["marker_2"]
    )
    data_frame = colocalization_data_frame.loc[~same_marker_mask, :]
    merged_differential_colocalization = pd.DataFrame()
    for target in targets:
        differential_colocalization = (
            data_frame.groupby(["marker_1", "marker_2"])
            .apply(
                lambda marker_data: wilcoxon_test(
                    marker_data,
                    reference=reference,
                    target=target,
                    contrast_column=contrast_column,
                    value_column=value_column,
                )
            )
            .reset_index()
        )

        # If a marker appears only in one of the datasets,
        # it's differential value will be NAN
        nan_values = differential_colocalization[
            differential_colocalization["median_difference"].isna()
        ].index
        differential_colocalization.drop(
            nan_values,
            axis="index",
            inplace=True,
        )

        _, pvals_corrected, *_ = multipletests(
            differential_colocalization["p_value"], method="bonferroni"
        )
        differential_colocalization["p_adj"] = pvals_corrected
        differential_colocalization["target"] = target
        merged_differential_colocalization = pd.concat(
            (merged_differential_colocalization, differential_colocalization), axis=0
        )

    return merged_differential_colocalization


class ColocalizationAnalysis(PerComponentAnalysis):
    """Run colocalization analysis on each component."""

    ANALYSIS_NAME = "colocalization"

    def __init__(
        self,
        transformation_type: TransformationTypes,
        neighbourhood_size: int,
        n_permutations: int,
        min_region_count: int,
        min_marker_count: int,
    ):
        """Initialize the ColocalizationAnalysis.

        :param transformation_type: transformation method to use
        :param neighbourhood_size: size of the neighbourhood to consider
        :param n_permutations: Select number of permutations used to
                               calculate empirical z-scores and p-values of the
                               colocalization values
        :param min_region_count: The minimum size of the region (e.g. number
                             of counts in the neighbourhood) required
                             for it to be considered for colocalization analysis
        :param min_marker_count: the minimum number of counts of a marker to calculate
                             colocalization
        """
        self.transformation_type = transformation_type
        self.neighbourhood_size = neighbourhood_size
        self.n_permutations = n_permutations
        self.min_region_count = min_region_count
        self.min_marker_count = min_marker_count

    def run_on_component(self, component: Graph, component_id: str) -> pd.DataFrame:
        """Run colocalization analysis on the component."""
        logger.debug("Running colocalization analysis on component %s", component_id)
        return colocalization_from_component_graph(
            graph=component,
            component_id=component_id,
            transformation=self.transformation_type,
            neighbourhood_size=self.neighbourhood_size,
            n_permutations=self.n_permutations,
            min_region_count=self.min_region_count,
            min_marker_count=self.min_marker_count,
        )

    def post_process_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Post process the colocalization data.

        This will adjust the p-values using the Benjamini-Hochberg method.
        """
        logger.debug("Post processing colocalization analysis data")
        if data.empty:
            return data
        p_value_columns = filter(lambda x: "_p" in x, data.columns)
        for p_value_col in p_value_columns:
            data.insert(
                data.columns.get_loc(p_value_col) + 1,
                f"{p_value_col}_adjusted",
                correct_pvalues(data[p_value_col].to_numpy()),
            )
        return data

    def add_to_pixel_dataset(
        self, data: pd.DataFrame, pxl_dataset: PixelDataset
    ) -> PixelDataset:
        """Add the colocalization data to the PixelDataset."""
        logger.debug("Adding colocalization analysis data to PixelDataset")
        pxl_dataset.colocalization = data
        return pxl_dataset



"""Module for handling commandline metadata.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import json
import logging
import typing
from pathlib import Path
from typing import Any, Mapping, Optional, TypeVar

import click
import pydantic

logger = logging.getLogger(__name__)
CommandOptionValue = TypeVar(
    "CommandOptionValue", None, str, int, float, bool, list[str]
)


# -----------------------------------------------------------------------
# Some helper functions to deal with commandline metadata
# -----------------------------------------------------------------------


def _clean_commmand_path(data: Mapping[str, Any]) -> str:
    """Clean up the command path from a parsed meta.json file.

    :param data: The parsed meta.json file
    :returns str: The cleaned command.
    """
    command_path = data["cli"]["command"].split(" ")
    return " ".join(command_path)


def _find_click_command(
    click_context: click.Group, data: Mapping[str, Any]
) -> click.Command:
    """Find the click command for a given parsed meta.json file."""
    command_path = data["cli"]["command"].split(" ")
    command_group: click.Group = click_context

    clean_command_path = _clean_commmand_path(data).split(" ")

    if len(command_path) <= 1:
        raise ValueError("Expected at least one subcommand")

    for subcommand in clean_command_path[1:-1]:
        if subcommand in command_group.commands:
            command_group = typing.cast(click.Group, command_group.commands[subcommand])
        else:
            raise ValueError(f"Unknown command {subcommand}")

    leaf_command = command_group.commands[clean_command_path[-1]]
    return leaf_command


def _process_meta_json_data(data: Mapping[str, Any]) -> CommandInfo:
    """Process a single metadata file and generate the parameter info object.

    :param data: The parsed meta.json file
    :return: A :class:`CommandInfo` object.
    """
    from pixelator.cli import main_cli

    leaf_command = _find_click_command(main_cli, data)
    param_data: list[CommandOption] = []
    opt_lookup = {p.opts[0]: p for p in leaf_command.params}
    clean_command_name = _clean_commmand_path(data)

    for param_name, param_value in data["cli"]["options"].items():
        option_info = opt_lookup.get(param_name)

        if option_info is None:
            logger.warning(
                f'Unknown parameter "{param_name}" for command: "{clean_command_name}"'
            )
            param_data.append(
                CommandOption(
                    name=param_name,
                    value=param_value,
                    default_value=None,
                    description=None,
                )
            )
            continue

        help_text = None
        if isinstance(option_info, click.Option):
            help_text = option_info.help

        param_data.append(
            CommandOption(
                name=param_name,
                value=param_value,
                default_value=option_info.default,
                description=help_text,
            )
        )

    command = CommandInfo(
        command=clean_command_name,
        options=param_data,
    )

    return command


class CommandOption(pydantic.BaseModel, typing.Generic[CommandOptionValue]):
    """Dataclass for passing command options/flags to qc report."""

    name: str = pydantic.Field(
        description="The name of the option. (e.g. `--min-reads`)"
    )

    value: CommandOptionValue = pydantic.Field(description="The value of the option.")
    default_value: Optional[CommandOptionValue] = pydantic.Field(
        description="The default value for this option."
    )
    description: str | None = pydantic.Field(
        description="The description of the option."
    )


class CommandInfo(pydantic.BaseModel):
    """Dataclass for passing all options of a command to qc report."""

    command: str = pydantic.Field(description="The name of the command.")

    options: list[CommandOption] = pydantic.Field(
        description="The list of options for the command."
    )

    @classmethod
    def from_json(cls, p: Path) -> CommandInfo:
        """Initialize an :class:`CommandInfo` from a json file.

        :param p: The path to the report file.
        :return: A :class:`CommandInfo` object.
        """
        with open(p, "r") as fh:
            file_data = json.load(fh)

        command_info = _process_meta_json_data(file_data)
        return command_info



"""Utility methods for the pixelator.pna package.

Copyright © 2024 Pixelgen Technologies AB.
"""

from .two_bit_encoding import pack_2bits as pack_2bits
from .two_bit_encoding import unpack_2bits as unpack_2bits
from .units import parse_size as parse_size
from .utils import *



"""QC report builder.

The QCReportBuilder is used to inject all QC report data into the
template and write the final QC report to a file.

Copyright © 2022 Pixelgen Technologies AB.
"""

import base64
import dataclasses
import gzip
import json
import logging
from pathlib import Path
from typing import Any, BinaryIO, ClassVar, Dict, List, Optional, Tuple, Union

import lxml.etree
import semver
from lxml.etree import _Element as LxmlElement
from lxml.html import builder as E

from pixelator.common.report.json_encoder import PixelatorJSONEncoder
from pixelator.mpx.report.qcreport.types import Metrics, QCReportData, SampleInfo

logger = logging.getLogger(__name__)

DEFAULT_QC_REPORT_TEMPLATE = Path(__file__).parent / "template.html"


class QCReportBuilder:
    """Build a qc report from a html template and the required data (CSV and JSON).

    This will parse a qc report template and, using the write method, inject CSV
    and JSON strings into the template using `html <script>` tags under the
    `<body>` tag.

    Tags are labeled with the custom attributes `data-type` and `data-dataset-id`
    to identify the data.

    Note that since we are creating a single sample report (for now) the
    `data-dataset-id` is always set to '0'.

    >>> builder = QCReportBuilder("./path/to/template.html")
    >>> with open("report.html", "wb") as fp:
    ...    builder.write(fp, sample_info, metrics, data)
    """

    _JSON_OPTIONS: ClassVar[Dict[str, Any]] = {"indent": None, "separators": (",", ":")}
    VERSIONS_CONSTRAINTS: ClassVar[List[str]] = ["<0.15.0", ">=0.10.0"]

    def __init__(self, template: Union[str, Path] = DEFAULT_QC_REPORT_TEMPLATE):
        """Construct a QC report builder given a html template.

        :param template: path to the QC report template
        :raises FileNotFoundError: if the template file does not exist
        """
        if not Path(template).exists():
            raise FileNotFoundError(
                f"QC Report template not found at specified location: {template}"
            )
        self.template = Path(template)

    def _load_template(self) -> Tuple[LxmlElement, LxmlElement]:
        """Load and parse the QC report template.

        :raises AssertionError: if no body tag is found in the template
        :raises AssertionError: if the QC report template version is not supported
        :rtype: Tuple[LxmlElement, LxmlElement]
        """
        logger.debug("Loading QC report template %s", self.template)

        # create a parser
        # Use huge_tree to allow large lxml objects and prevent
        # silent truncation of the html output.
        parser = lxml.etree.HTMLParser(huge_tree=True)

        # load and parse QC report template
        document = None
        with open(self.template, "rb") as f:
            document = lxml.html.parse(f, parser)

        root = document.getroot()
        body = document.find("body")

        if body is None:
            raise AssertionError("Could not find body element in QC report template")

        version = root.xpath('//meta[@name="application-name"]/@data-version')
        version = version[0] if version else ""
        self._check_version_compatibility(version)

        logger.debug("QC report template %s loaded", self.template)
        return root, body

    def _check_version_compatibility(self, version: str):
        for constraint in self.VERSIONS_CONSTRAINTS:
            check = semver.VersionInfo.parse(version).match(constraint)
            if not check:
                raise AssertionError(
                    f"Unsupported QC report version. Version {version} does "
                    f"not satisfy constraint: {constraint}"
                )

    def write(  # noqa: DOC502
        self,
        fp: BinaryIO,
        sample_info: SampleInfo,
        metrics: Metrics,
        data: QCReportData,
        metrics_definition_file: Optional[Path] = None,
    ) -> None:
        """Inject given data into the QC report and write the results to a stream.

        :param fp: binary stream to write the report to
        :param sample_info: Sample information
        :param metrics: Metrics for the sample
        :param data: Data for dynamic figures
        :param metrics_definition_file: Path to the metrics definition file
        :raises: AssertionError if not body tag is found in the template
        """
        root, body = self._load_template()

        # create the necessary script elements
        elements = []

        if metrics_definition_file is not None:
            elements.append(
                self._build_metric_definition_file_element(
                    metrics_definition_file, body
                )
            )

        elements.append(self._build_sample_and_metrics_element(sample_info, metrics))

        if data.ranked_component_size is not None:
            elements.append(
                self._build_ranked_component_size_element(data.ranked_component_size)
            )

        if data.component_data is not None:
            elements.append(self._build_component_data_element(data.component_data))

        if data.antibodies_per_cell is not None:
            elements.append(
                self._build_antibodies_per_component_element(data.antibodies_per_cell)
            )

        if data.sequencing_saturation is not None:
            elements.append(
                self._build_sequencing_saturation_element(data.sequencing_saturation)
            )

        if data.antibody_percentages is not None:
            elements.append(
                self._build_antibody_percentages_element(data.antibody_percentages)
            )

        if data.antibody_counts is not None:
            elements.append(self._build_antibody_counts_element(data.antibody_counts))

        if data.reads_per_molecule_frequency is not None:
            elements.append(
                self._build_reads_per_molecule_frequency_element(
                    data.reads_per_molecule_frequency
                )
            )

        # inject the elements in the body (in order for a more structured html)
        for idx, el in enumerate(elements):
            body.insert(idx, el)

        # write to html
        lxml.etree.ElementTree(root).write(fp, method="html", encoding="utf-8")

        logger.debug("Data injected in QC report template %s", self.template)

    def _build_metric_definition_file_element(
        self, metrics_definition_file: Path, template_body: LxmlElement
    ) -> LxmlElement:
        """Create a lxml HTML object to inject the metrics definition file."""
        template_body.cssselect('script[data-type="metric-definitions"]')
        if len(template_body) > 0:
            metrics_definition_file_el = template_body[0]
            parent = metrics_definition_file_el.getparent()
            if parent:
                parent.remove(metrics_definition_file_el)

        metrics_definition_file_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "metric-definitions",
            }
        )

        with open(metrics_definition_file, "r") as f:
            data = f.read()

        metrics_definition_file_el.text = self._compress_data(data)
        return metrics_definition_file_el

    def _build_sample_and_metrics_element(
        self, sample_info: SampleInfo, metrics: Metrics
    ) -> LxmlElement:
        """Create a lxml HTML object to inject the metrics and sample info."""
        metrics_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "metrics",
                "data-dataset-id": "0",
            }
        )

        combined = {
            "info": dataclasses.asdict(sample_info),
            "metrics": metrics,
        }
        data = json.dumps(combined, **self._JSON_OPTIONS, cls=PixelatorJSONEncoder)
        metrics_el.text = self._compress_data(data)
        return metrics_el

    def _build_ranked_component_size_element(self, data: str) -> LxmlElement:
        """Create a lxml HTML injecting the ranked component data.

        This data is used for the rank plot and the component size vs marker
        scatter plot.
        """
        ranked_component_size_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "ranked-component-size",
                "data-dataset-id": "0",
            }
        )
        ranked_component_size_el.text = self._compress_data(data)
        return ranked_component_size_el

    def _build_component_data_element(self, data: str) -> LxmlElement:
        """Create a lxml HTML injecting the component data."""
        component_data_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "component-data",
                "data-dataset-id": "0",
            }
        )
        component_data_el.text = self._compress_data(data)
        return component_data_el

    def _build_antibodies_per_component_element(self, data: str) -> LxmlElement:
        """Create a lxml HTML injecting the antibodies_per_cell data."""
        antibodies_per_cell_el = E.SCRIPT(
            **{
                "type": "text/csv",
                "data-type": "antibodies-per-component",
                "data-dataset-id": "0",
            }
        )
        antibodies_per_cell_el.text = data
        return antibodies_per_cell_el

    def _build_sequencing_saturation_element(self, data: str):
        """Create an HTML object injecting the sequencing saturation data."""
        sequencing_saturation_el = E.SCRIPT(
            **{
                "type": "text/csv",
                "data-type": "sequencing-saturation",
                "data-dataset-id": "0",
            }
        )
        sequencing_saturation_el.text = data
        return sequencing_saturation_el

    def _build_antibody_percentages_element(self, data: str) -> LxmlElement:
        """Create an HTML object injecting the antibody counts data."""
        antibody_counts_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "antibody-percentages",
                "data-dataset-id": "0",
            }
        )
        antibody_counts_el.text = self._compress_data(data)
        return antibody_counts_el

    def _build_antibody_counts_element(self, data: str) -> LxmlElement:
        """Create an HTML object injecting the antibody counts data."""
        antibody_distribution_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "antibody-counts",
                "data-dataset-id": "0",
            }
        )
        antibody_distribution_el.text = self._compress_data(data)
        return antibody_distribution_el

    def _build_reads_per_molecule_frequency_element(self, data: str) -> LxmlElement:
        """Create an HTML object injecting the data for the reads per umi histogram."""
        reads_per_molecule_frequency_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "reads-per-molecule-frequency",
                "data-dataset-id": "0",
            }
        )
        reads_per_molecule_frequency_el.text = self._compress_data(data)
        return reads_per_molecule_frequency_el

    @staticmethod
    def _compress_data(data: str):
        """Compress the data using gzip and encode with base64."""
        return base64.b64encode(gzip.compress(data.encode("utf-8")))



"""
Plugin helpers for the cli.

Copyright © 2022 Pixelgen Technologies AB.
"""

import importlib.metadata

try:
    from importlib.metadata import EntryPoints
except ImportError:
    # Python 3.8 and 3.9
    pass

import logging
from importlib.metadata import EntryPoint
from typing import Generator, List, Union

from click import Group

logger = logging.getLogger(__name__)


def fetch_cli_plugins() -> Generator[EntryPoint, None, None]:
    """
    Find plugins and return them as in a generator.

    :yields EntryPoint: The entrypoint object
    :returns: A generator of the EntryPoint objects
    """
    eps = importlib.metadata.entry_points()
    group = "pixelator.cli_plugin"
    selected_entrypoints: Union[List[EntryPoint], EntryPoints]

    if hasattr(eps, "select"):
        # New interface in Python 3.10 and newer versions of the
        # importlib_metadata backport.
        selected_entrypoints = eps.select(group=group)
    else:
        # Older interface, deprecated in Python 3.10 and recent
        # importlib_metadata, but we need it in Python 3.8 and 3.9.
        selected_entrypoints = eps.get(group, [])  # type: ignore

    for entrypoint in selected_entrypoints:
        logger.debug("Detected cli plugin %s", entrypoint.name)
        yield entrypoint


def add_cli_plugins(group: Group) -> None:
    """Add all cli plugins we can find to the provided group.

    :param group: An instance of `click.Group` to add sub commands to
    :returns: None
    """
    for entrypoint in fetch_cli_plugins():
        logger.debug("Loading cli plugin %s", entrypoint.name)
        group.add_command(entrypoint.load())



"""Copyright © 2023 Pixelgen Technologies AB."""

from __future__ import annotations

import enum
import json
import typing
from collections import defaultdict
from typing import Iterable, Optional

from pixelator.mpx.report.common.workdir import SingleCellStageLiteral
from pixelator.mpx.report.models import CommandInfo, CommandOption
from pixelator.pna.report.common.workdir import (
    SingleCellPNAStage,
    SingleCellPNAStageLiteral,
)

CommandInfoDict: typing.TypeAlias = dict[str, CommandInfo]
CommandOptionDict: typing.TypeAlias = dict[str, CommandOption]
CommandIndexTuple: typing.TypeAlias = tuple[
    dict[str, CommandInfo], dict[str, CommandOptionDict]
]

_SINGLE_CELL_STAGES_TO_CACHE_KEY_MAPPING: dict[
    SingleCellStageLiteral | SingleCellPNAStageLiteral, str | list[str]
] = {
    "amplicon": [
        "pixelator single-cell-pna amplicon",
    ],
    "demux": [
        "pixelator single-cell-pna demux",
    ],
    "collapse": [
        "pixelator single-cell-pna collapse",
    ],
    "graph": ["pixelator single-cell-pna graph"],
    "layout": [
        "pixelator single-cell-pna layout",
    ],
    "analysis": [
        "pixelator single-cell-pna analysis",
    ],
    "report": [
        "pixelator single-cell-pna report",
    ],
}


class CLIInvocationInfo:
    """List of commandline invocations from a pixelator working directory.

    :ivar sample_id: The sample_id on which these commands where run
    """

    def __init__(self, data: list[CommandInfo], sample_id: Optional[str]):
        """Initialize the CLIInvocationInfo object.

        :param data: The commandline invocations
        :param sample_id: The sample_id on which these commands where run
        """
        self.sample_id = sample_id
        self._data = data
        self._commands_index, self._options_index = self._index_parameter_info(data)

    def __iter__(self) -> typing.Iterator[CommandInfo]:
        """Iterate over the commandline invocations."""
        return iter(self._data)

    def __len__(self) -> int:
        """Return the number of commandline invocations."""
        return len(self._data)

    def __getitem__(self, index) -> CommandInfo:
        """Return the commandline invocation at the given index."""
        return self._data[index]

    @staticmethod
    def _index_parameter_info(data: list[CommandInfo]) -> CommandIndexTuple:
        """Create two lookup tables for querying parameter info.

        :param data: the result from `generate_parameter_info`
        :returns: A tuple with two lookup tables
        :rtype: Tuple[dict[str, CommandInfo], dict[str, dict[str, CommandOption]]]
        """
        command_index = {}
        command_option_index: dict[str, dict[str, CommandOption]] = defaultdict(dict)

        for command_info in data:
            command_index[command_info.command] = command_info
            for option in command_info.options:
                command_option_index[command_info.command][option.name] = option

        return command_index, command_option_index

    def get_stage(
        self, stage: SingleCellPNAStage | SingleCellPNAStageLiteral
    ) -> CommandInfo | None:
        """Return the CommandInfo object for a given stage.

        If no commandline metadata is found for the stage, None is returned.

        :param stage: The stage to return the commandline for
        :returns: The commandline info for given stage
        """
        stage = stage.value if isinstance(stage, enum.Enum) else stage
        if stage in _SINGLE_CELL_STAGES_TO_CACHE_KEY_MAPPING:
            stage_key = _SINGLE_CELL_STAGES_TO_CACHE_KEY_MAPPING[stage]
        else:
            stage_key = stage

        res = None

        if isinstance(stage_key, list):
            for key in stage_key:
                res = self._commands_index.get(key)
                if res:
                    break
        else:
            res = self._commands_index.get(stage_key)

        return res

    def get_option(
        self, stage: SingleCellPNAStage | SingleCellPNAStageLiteral, option: str
    ) -> CommandOption:
        """Return the commandline options used to invoke a pixelator command.

        :param stage: The stage to return the commandline for
        :param option: The option to return. eg. "--design"
        :raises KeyError:
            If no commandline metadata is found for the stage,
            or the option does not exist for that stage.
        :returns CommandOption: The commandline option for the stage.
        """
        stage = stage.value if isinstance(stage, enum.Enum) else stage

        if stage in _SINGLE_CELL_STAGES_TO_CACHE_KEY_MAPPING:
            stage_key = _SINGLE_CELL_STAGES_TO_CACHE_KEY_MAPPING[stage]
        else:
            stage_key = stage

        stage_dict = None

        if isinstance(stage_key, list):
            for key in stage_key:
                stage_dict = self._options_index.get(key)
                if stage_dict:
                    break

        else:
            stage_dict = self._options_index.get(stage_key)

        if stage_dict is None:
            stage_key = stage_key[0] if isinstance(stage_key, list) else stage_key
            raise KeyError(f"No commandline metadata found for stage: {stage_key}")

        return stage_dict[option]

    def to_json(self, **json_options) -> str:
        """Return a JSON formatted string of the commandline invocations.

        :param json_options: Additional options to pass to json.dumps
        """
        return json.dumps(
            [v.model_dump(mode="json") for v in self._data], **json_options
        )



"""Module for estimating colocalization statistics.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
from typing import Callable, Iterable, Optional, Tuple

import numpy as np
import pandas as pd
from scipy.stats import norm

from pixelator.mpx.analysis.colocalization.types import (
    CoLocalizationFunction,
)
from pixelator.mpx.analysis.types import RegionByCountsDataFrame

logger = logging.getLogger(__name__)


def estimate_observation_statistics(
    observations: pd.DataFrame,
    permutation_results: pd.DataFrame,
    funcs: Tuple[CoLocalizationFunction, ...],
):
    """Estimates the observation statistics for colocalization analysis.

    :param observations: The observed data.
    :param permutation_results: The permutation results.
    :param funcs: Tuple of colocalization functions.
    :return: The estimated observation statistics.
    :rtype: pd.DataFrame
    """

    def estimates():
        for func in funcs:
            func_name = func.name
            # Since we are using a normal distribution we can just calculate the
            # mean and stdev directly
            permutation_stats = permutation_results.groupby(
                ["marker_1", "marker_2"]
            ).agg(
                **{
                    f"{func_name}_mean": pd.NamedAgg(f"{func_name}_perm", "mean"),
                    f"{func_name}_stdev": pd.NamedAgg(f"{func_name}_perm", "std"),
                }
            )

            merged = pd.merge(
                observations[func_name], permutation_stats, on=["marker_1", "marker_2"]
            )
            merged[f"{func_name}_z"] = (
                merged[func_name] - merged[f"{func_name}_mean"]
            ) / merged[f"{func_name}_stdev"]
            merged[f"{func_name}_p_value"] = norm.sf(np.abs(merged[f"{func_name}_z"]))
            yield merged

    result = pd.concat(estimates(), axis=1)
    return result.reset_index()


def permutation_analysis_results(
    data: RegionByCountsDataFrame,
    funcs: Tuple[CoLocalizationFunction, ...],
    permuter: Callable[
        [RegionByCountsDataFrame, int], Iterable[RegionByCountsDataFrame]
    ],
    transformer: Optional[
        Callable[[RegionByCountsDataFrame], RegionByCountsDataFrame]
    ] = None,
    n: int = 50,
    random_seed: Optional[int] = None,
) -> pd.DataFrame:
    """Perform permutation analysis on colocalization data.

    :param data: The input data for colocalization analysis.
    :param funcs: A tuple of colocalization functions to apply.
    :param permuter: A function that generates permuted data for analysis.
    :param transformer: A function to transform the permuted data before analysis. (optional)
    :param n: The number of permutations to perform. Default is 50.
    :param random_seed: The random seed for reproducibility. Default is None.
    :return: The results of the permutation analysis.
    :rtype: pd.DataFrame
    """

    def constuct_permutation_data(data, n):
        for idx, permuted_df in enumerate(permuter(data, n=n, random_seed=random_seed)):
            df_for_comp = transformer(permuted_df) if transformer else permuted_df
            res = pd.concat(
                [
                    func.func(df_for_comp).rename(
                        columns={func.name: f"{func.name}_perm"}
                    )
                    for func in funcs
                ],
                axis=1,
            )
            res["permutation"] = idx
            res.set_index(["permutation"], inplace=True, append=True)
            res.index.names = ["marker_1", "marker_2", "permutation"]
            yield res

    permutations = pd.concat(constuct_permutation_data(data, n))
    return permutations



"""Module for PixelDataset backends.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

from functools import cached_property
from typing import (
    Any,
    Dict,
    Optional,
    Protocol,
)

import pandas as pd
import polars as pl
from anndata import AnnData

from pixelator.common.types import PathType
from pixelator.mpx.pixeldataset.datastores import PixelDataStore
from pixelator.mpx.pixeldataset.precomputed_layouts import PreComputedLayouts
from pixelator.mpx.pixeldataset.utils import (
    _enforce_edgelist_types,
)


class PixelDatasetBackend(Protocol):
    """Protocol for PixelDataset backends.

    This protocol defines the methods required by a PixelDataset backend.
    Any class that fulfills this contract can be used as a backend for a
    PixelDataset.

    Normally the backend used by pixelator is a .pxl file, but by implementing
    this protocol you can provide other backends, to for example load data directly
    from a database.
    """

    @property
    def adata(self) -> AnnData:
        """Return the AnnData instance."""

    @adata.setter
    def adata(self, value: AnnData) -> None:
        """Set the AnnData instance."""

    @property
    def edgelist(self) -> pd.DataFrame:
        """Return the edgelist."""

    @edgelist.setter
    def edgelist(self, value: pd.DataFrame) -> None:
        """Set the edge list instance."""

    @property
    def edgelist_lazy(self) -> pl.LazyFrame:
        """Return the edgelist as a LazyFrame."""

    @property
    def polarization(self) -> Optional[pd.DataFrame]:
        """Return the polarization data."""

    @polarization.setter
    def polarization(self, value: pd.DataFrame) -> None:
        """Set the polarization data frame."""

    @property
    def colocalization(self) -> Optional[pd.DataFrame]:
        """Return the colocalization data."""

    @colocalization.setter
    def colocalization(self, value: pd.DataFrame) -> None:
        """Set the colocalization data frame."""

    @property
    def metadata(self) -> Optional[Dict]:
        """Return the metadata object."""

    @metadata.setter
    def metadata(self, value: Dict) -> Optional[Dict]:
        """Set the metadata object."""

    @property
    def precomputed_layouts(self) -> PreComputedLayouts:
        """Get the precomputed layouts for the component graphs.

        Please note that since these have been pre-computed, if you have made
        changes to the underlying data, the layout might not be valid anymore.
        """
        ...

    @precomputed_layouts.setter
    def precomputed_layouts(self, value: PreComputedLayouts | None) -> None:
        """Set the precomputed layouts for the component graphs."""
        ...


class ObjectBasedPixelDatasetBackend(PixelDatasetBackend):
    """A backend for PixelDataset that is backed by in memory objects.

    `ObjectBasedPixelDatasetBackend` provides a backend for PixelDatasets that
    are stored entirely in memory. This is mostly used by Pixelator internally
    the first time the PixelDataset instance is created.

    Note that it will make it's own copy of all the provided input data.
    """

    def __init__(
        self,
        adata: AnnData,
        edgelist: pd.DataFrame,
        metadata: Optional[Dict[str, Any]] = None,
        polarization: Optional[pd.DataFrame] = None,
        colocalization: Optional[pd.DataFrame] = None,
        precomputed_layouts: Optional[PreComputedLayouts] = None,
        copy: bool = True,
        allow_edgelist_to_be_empty: bool = False,
    ) -> None:
        """Create a new instance of ObjectBasedPixelDatasetBackend.

        :param adata: an AnnData instance
        :param edgelist: an edgelist dataframe
        :param metadata: a dict with metadata, defaults to None
        :param polarization: a polarization dataframe, defaults to None
        :param colocalization: a colocalization dataframe, defaults to None
        :param copy: decide if the input data should be copied or not. Defaults to True.
        :param allow_edgelist_to_be_empty: allow the edgelist to be empty.
                                           Defaults to False.
        :raises AssertionError: if `adata` or `edgelist` contains no data.
        """
        if adata is None or adata.n_obs == 0:
            raise AssertionError("adata cannot be empty")

        if edgelist is None or edgelist.shape[0] == 0:
            if not allow_edgelist_to_be_empty:
                raise AssertionError("edgelist cannot be empty")
            edgelist = _enforce_edgelist_types(pd.DataFrame())

        self._edgelist = _enforce_edgelist_types(edgelist.copy() if copy else edgelist)
        self._adata = adata.copy() if copy else adata
        self._metadata = metadata

        self._polarization = None
        if polarization is not None:
            self._polarization = polarization.copy() if copy else polarization

        self._colocalization = None
        if colocalization is not None:
            self._colocalization = colocalization.copy() if copy else colocalization

        self._precomputed_layouts = PreComputedLayouts.create_empty()
        if precomputed_layouts is not None:
            self._precomputed_layouts = precomputed_layouts

    @property
    def adata(self) -> AnnData:
        """Get the AnnData object for the pixel dataset."""
        return self._adata

    @adata.setter
    def adata(self, value: AnnData) -> None:
        """Set the AnnData object for the pixel dataset."""
        self._adata = value

    @property
    def edgelist(self) -> pd.DataFrame:
        """Get the edge list for the pixel dataset."""
        return self._edgelist

    @edgelist.setter
    def edgelist(self, value: pd.DataFrame) -> None:
        """Set the edge list for the pixel dataset."""
        self._edgelist = value

    @property
    def edgelist_lazy(self) -> pl.LazyFrame:
        """Get a lazy frame representation of the edgelist."""
        return pl.LazyFrame(self._edgelist)

    @property
    def metadata(self) -> Optional[pd.DataFrame]:
        """Get the metadata for the pixel dataset."""
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]) -> None:
        """Set the metadata for the pixel dataset."""
        self._metadata = value

    @property
    def polarization(self) -> Optional[pd.DataFrame]:
        """Get the polarization scores for the pixel dataset."""
        return self._polarization

    @polarization.setter
    def polarization(self, value: pd.DataFrame) -> None:
        """Set the polarization scores for the pixel dataset."""
        self._polarization = value

    @property
    def colocalization(self) -> Optional[pd.DataFrame]:
        """Get the co-localization scores for the pixel dataset."""
        return self._colocalization

    @colocalization.setter
    def colocalization(self, value: pd.DataFrame) -> None:
        """Set the co-localization scores for the pixel dataset."""
        self._colocalization = value

    @property
    def precomputed_layouts(self) -> PreComputedLayouts:
        """Get the precomputed layouts."""
        if self._precomputed_layouts is None:
            return PreComputedLayouts.create_empty()
        return self._precomputed_layouts

    @precomputed_layouts.setter
    def precomputed_layouts(self, value: PreComputedLayouts | None) -> None:
        """Set the precomputed layouts."""
        if value is None:
            self._precomputed_layouts = PreComputedLayouts.create_empty()
        self._precomputed_layouts = value


class FileBasedPixelDatasetBackend(PixelDatasetBackend):
    """A file based backend for PixelDataset.

    `FileBasedPixelDatasetBackend` is used to lazily fetch information from
    a .pxl file. Once the file has been read the data will be cached
    in memory.
    """

    def __init__(
        self, path: PathType, datastore: Optional[PixelDataStore] = None
    ) -> None:
        """Create a filebased backend instance.

        Create a backend, fetching information from the .pxl file
        in `path`.

        :param path: Path to the .pxl file
        """
        self._path = path
        if not datastore:
            datastore = PixelDataStore.guess_datastore_from_path(path)
        self._datastore = datastore
        self._precomputed_layouts: PreComputedLayouts | None = None

    @cached_property
    def adata(self) -> AnnData:
        """Get the AnnData object for the pixel dataset."""
        return self._datastore.read_anndata()

    @cached_property
    def edgelist(self) -> pd.DataFrame:
        """Get the edge list object for the pixel dataset."""
        return self._datastore.read_edgelist()

    @property
    def edgelist_lazy(self) -> pl.LazyFrame:
        """Get a lazy frame representation of the edgelist."""
        return self._datastore.read_edgelist_lazy()

    @cached_property
    def polarization(self) -> Optional[pd.DataFrame]:
        """Get the polarization object for the pixel dataset."""
        return self._datastore.read_polarization()

    @cached_property
    def colocalization(self) -> Optional[pd.DataFrame]:
        """Get the colocalization object for the pixel dataset."""
        return self._datastore.read_colocalization()

    @cached_property
    def metadata(self) -> Optional[Dict]:
        """Get the metadata object for the pixel dataset."""
        return self._datastore.read_metadata()

    @property
    def precomputed_layouts(self) -> PreComputedLayouts:
        """Get the precomputed layouts."""
        # If it is None it means it is uninitialized, and we should
        # attempt to read it lazily

        if isinstance(self._precomputed_layouts, PreComputedLayouts):
            return self._precomputed_layouts

        if self._precomputed_layouts is None:
            self._precomputed_layouts = self._datastore.read_precomputed_layouts()
        return self._precomputed_layouts

    @precomputed_layouts.setter
    def precomputed_layouts(self, value: PreComputedLayouts | None) -> None:
        """Set the precomputed layouts."""
        if value is None:
            self._precomputed_layouts = PreComputedLayouts.create_empty()
            return
        self._precomputed_layouts = value



"""Copyright © 2023 Pixelgen Technologies AB."""

from pixelator.common.config.config_class import (
    Config,
    load_assays_package,
    load_panels_package,
)
from pixelator.mpx.config.plugin import load_config_plugins

config = Config()

assays_package = "pixelator.mpx.resources.assays"
config = load_assays_package(config, assays_package)

panels_package = "pixelator.mpx.resources.panels"
config = load_panels_package(config, panels_package)

config = load_config_plugins(config)



"""Functions and classes related to reporting for the collapse stage.

Copyright © 2024 Pixelgen Technologies AB
"""

import pydantic

from pixelator.pna.collapse.independent.collapser import MarkerCorrectionStats
from pixelator.pna.report.models.base import SampleReport


class IndependentCollapseSampleReport(SampleReport):
    """Model for report data returned by the collapse stage."""

    report_type: str = "collapse-independent"

    input_reads: int = pydantic.Field(
        ..., description="The number of input reads processed."
    )

    input_molecules: int = pydantic.Field(
        ...,
        description="The number of molecules (unique input reads) processed.",
    )

    output_molecules: int = pydantic.Field(
        ..., description="The total number of error-corrected molecules detected."
    )

    @pydantic.computed_field(  # type: ignore
        description="The number of output reads.", return_type=int
    )
    @property
    def output_reads(self) -> int:
        """The total number of error-corrected output reads.

        This is an alias for the input reads as no reads are removed during the collapse step.
        """
        return self.input_reads

    @pydantic.computed_field(  # type: ignore
        description="The number of UMI1 reads that had error and were corrected.",
        return_type=int,
    )
    @property
    def corrected_umi1_reads(self) -> int:
        """The number of UMI1 reads that had an error and were corrected."""
        return sum([m.corrected_reads for m in self.markers if m.region_id == "umi-1"])

    @pydantic.computed_field(  # type: ignore
        description="The number of UMI2 reads that had error and were corrected.",
        return_type=int,
    )
    @property
    def corrected_umi2_reads(self) -> int:
        """The number of UMI2 reads that had an error and were corrected."""
        return sum([m.corrected_reads for m in self.markers if m.region_id == "umi-2"])

    corrected_reads: int = pydantic.Field(
        ...,
        description="The number of reads that had an error in either UMI1, UMI2 or both and were corrected.",
    )

    markers: list[MarkerCorrectionStats] = pydantic.Field(
        ..., description="The statistics for each marker pair."
    )



"""Constants used as part of the annotation modules.

Copyright © 2022 Pixelgen Technologies AB.
"""

MINIMUM_NBR_OF_CELLS_FOR_ANNOTATION = 20

# size limit constants
MINIMUM_N_EDGES_CELL_SIZE = 300
MINIMUM_NBR_OF_CELLS_FOR_SIZE_LIMIT = 20
CELL_MIN_SIZE_SMOOTHING_FACTOR = 0.8
CELL_MAX_SIZE_SMOOTHING_FACTOR = 0
DISTANCE_DEVIATION_FACTOR = 5
# this is number of edges
PRE_FILTER_LIMIT = 20

TAU_HARD_THRESHOLD = 0.995
TAU_IQR_UPPER_THRESHOLD = 2
TAU_IQR_LOWER_THRESHOLD = 5



"""Copyright © 2023 Pixelgen Technologies AB."""

from pixelator.common.report.models import SummaryStatistics

from .adapterqc import AdapterQCSampleReport
from .amplicon import AmpliconSampleReport
from .analysis import AnalysisSampleReport
from .annotate import AnnotateSampleReport
from .collapse import CollapseSampleReport
from .commands import CommandInfo, CommandOption
from .demux import DemuxSampleReport
from .graph import GraphSampleReport
from .molecules_flow import MoleculesDataflowReport
from .preqc import PreQCSampleReport
from .reads_flow import ReadsDataflowReport

__all__ = [
    "AdapterQCSampleReport",
    "AmpliconSampleReport",
    "CollapseSampleReport",
    "DemuxSampleReport",
    "GraphSampleReport",
    "PreQCSampleReport",
    "AnalysisSampleReport",
    "AnnotateSampleReport",
    "ReadsDataflowReport",
    "MoleculesDataflowReport",
    "CommandInfo",
    "CommandOption",
    "SummaryStatistics",
]



"""Functions for creating plots that are useful with MPX data.

Copyright © 2023 Pixelgen Technologies AB.
"""

import warnings
from typing import Optional, Tuple

import anndata
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.colors import LinearSegmentedColormap, Normalize
from matplotlib.patches import Rectangle
from scipy.stats import gaussian_kde

from pixelator.mpx.plot.constants import Color
from pixelator.mpx.plot.layout_plots import (
    plot_2d_graph,
    plot_3d_from_coordinates,
    plot_3d_graph,
    plot_3d_heatmap,
)
from pixelator.mpx.plot.spatial_analysis_plots import (
    plot_colocalization_diff_heatmap,
    plot_colocalization_diff_volcano,
    plot_colocalization_heatmap,
    plot_polarity_diff_volcano,
)

sns.set_style("whitegrid")
jet_colormap = LinearSegmentedColormap.from_list("jet_colormap", Color.JETSET)


def scatter_umi_per_upia_vs_tau(
    data: pd.DataFrame, group_by: Optional[str] = None
) -> Tuple[plt.Figure, plt.Axes]:
    """Create a scatter plot of pixel content vs marker specificity (Tau).

    :param data: a pandas DataFrame with the columns 'umi_per_upia', 'tau', and 'tau_type'.
    :param group_by: a column in the DataFrame to group the plot by.

    :return: a scatter plot of pixel content vs marker specificity (Tau).
    :rtype: Tuple[plt.Figure, plt.Axes]
    :raises: ValueError if the required columns are not present in the DataFrame
    :raises: AssertionError if the data types are invalid
    """
    # Validate data
    required_columns = ["umi_per_upia", "tau", "tau_type"]
    if not all(col in data.columns for col in required_columns):
        raise ValueError(
            "'umi_per_upia', 'tau' and 'tau_type' must be available in the DataFrame"
        )
    if not all(
        [
            data["umi_per_upia"].dtype == "float64",
            data["tau"].dtype == "float64",
            data["tau_type"].dtype in ["object", "category"],
        ]
    ):
        raise AssertionError("Invalid data types")

    if group_by is not None:
        if group_by not in data.columns:
            raise ValueError(f"'{group_by}' is missing")
        assert data[group_by].dtype in [
            "object",
            "category",
        ], "'group_by' must be a character or factor"

    # Define palette
    palette = {
        "high": Color.ORANGERED2,
        "low": Color.SKYBLUE3,
        "normal": Color.LIGHTGREY,
    }

    # create plot
    grid = (
        sns.FacetGrid(
            data,
            col=group_by,
            hue="tau_type",
            palette=palette,
            height=4,
        )
        if group_by is not None
        else sns.FacetGrid(data, hue="tau_type", palette=palette, height=4)
    )
    grid.map(sns.scatterplot, "tau", "umi_per_upia")
    grid.set(yscale="log")
    grid.set_axis_labels("Marker specificity (Tau)", "Pixel content (UMI/UPIA)")
    grid.add_legend(title="Tau Type")

    return plt.gcf(), plt.gca()


def molecule_rank_plot(
    data: pd.DataFrame, group_by: Optional[str] = None
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot the number of molecules per component against its molecule rank.

    :param data: a pandas DataFrame with a column 'molecules' containing edge counts for MPX
    components.
    :param group_by: a column in the DataFrame to group the plot by.

    :return: a plot showing the number of molecules per component against its edge rank used
    for quality control.
    :rtype: Tuple[plt.Figure, plt.Axes]
    :raises: AssertionError if the required column(s) are not present in the DataFrame
    :raises: ValueError if the data types are invalid
    """
    if "molecules" not in data.columns and "edges" in data.columns:
        data["molecules"] = data["edges"]
    assert "molecules" in data.columns, "column 'molecules' is missing from DataFrame"
    assert (
        isinstance(data["molecules"], pd.Series) and data["molecules"].dtype == int
    ), "'molecules' must be a vector of integer values"

    if group_by is not None:
        assert group_by in data.columns, (
            f"group variable '{group_by}' not found in DataFrame"
        )

        if data[group_by].dtype not in ["object", "category"]:
            raise ValueError(
                f"Invalid class '{data[group_by].dtype}' for column '{group_by}'. "
                f"Expected a string or categorical value"
            )
        else:
            edge_rank_df = data[[group_by, "molecules"]].copy()
            edge_rank_df["rank"] = edge_rank_df.groupby([group_by])["molecules"].rank(
                ascending=False, method="first"
            )
    else:
        edge_rank_df = data[["molecules"]].copy()
        edge_rank_df["rank"] = edge_rank_df["molecules"].rank(
            ascending=False, method="first"
        )

    (
        sns.relplot(
            data=edge_rank_df, x="rank", y="molecules", hue=group_by, aspect=1.6
        )
        .set(xscale="log", yscale="log")
        .set_xlabels("Component rank (by number of molecules)")
        .set_ylabels("Number of molecules")
    )

    return plt.gcf(), plt.gca()


def edge_rank_plot(
    data: pd.DataFrame, group_by: Optional[str] = None
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot the number of edges per component against its edge rank.

    :param data: a pandas DataFrame with a column 'edges' containing edge counts for MPX
    components.
    :param group_by: a column in the DataFrame to group the plot by.

    :return: a plot showing the number of edges per component against its edge rank used
    for quality control.
    :rtype: Tuple[plt.Figure, plt.Axes]
    :raises: AssertionError if the required column(s) are not present in the DataFrame
    :raises: ValueError if the data types are invalid
    """
    warnings.warn(
        "edge_rank_plot is deprecated and will be removed in a future version. Use molecule_rank_plot instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    assert "edges" in data.columns, "column 'edges' is missing from DataFrame"
    assert isinstance(data["edges"], pd.Series) and data["edges"].dtype == int, (
        "'edges' must be a vector of integer values"
    )

    if group_by is not None:
        assert group_by in data.columns, (
            f"group variable '{group_by}' not found in DataFrame"
        )

        if data[group_by].dtype not in ["object", "category"]:
            raise ValueError(
                f"Invalid class '{data[group_by].dtype}' for column '{group_by}'. "
                f"Expected a string or categorical value"
            )
        else:
            edge_rank_df = data[[group_by, "edges"]].copy()
            edge_rank_df["rank"] = edge_rank_df.groupby([group_by])["edges"].rank(
                ascending=False, method="first"
            )
    else:
        edge_rank_df = data[["edges"]].copy()
        edge_rank_df["rank"] = edge_rank_df["edges"].rank(
            ascending=False, method="first"
        )

    (
        sns.relplot(data=edge_rank_df, x="rank", y="edges", hue=group_by, aspect=1.6)
        .set(xscale="log", yscale="log")
        .set_xlabels("Component rank (by number of edges)")
        .set_ylabels("Number of edges")
    )

    return plt.gcf(), plt.gca()


def cell_count_plot(
    data: pd.DataFrame,
    color_by: str,
    group_by: Optional[str] = None,
    flip_axes: bool = False,
) -> Tuple[plt.Figure, plt.Axes]:
    """Create a bar plot showing the component counts by group(s).

    :param data: a pandas DataFrame with group variable(s).
    :param color_by: a column in the DataFrame to color the bars by. This will be used as
    the group variable if `group_by` is not provided.
    :param: group_by: a column in the DataFrame to group the bars by.

    :return: a bar plot with component counts by group(s).
    :rtype: Tuple[plt.Figure, plt.Axes]
    :raises: ValueError if the required grouping variables are missing in the DataFrame
    :raises: AssertionError if the data types are invalid
    """
    # Validate inputs
    if group_by is not None:
        if group_by not in data.columns:
            raise ValueError(f"'{group_by}' is missing from DataFrame")

        if not pd.api.types.is_categorical_dtype(
            data[group_by]
        ) and not pd.api.types.is_string_dtype(data[group_by]):
            raise ValueError("'group_by' must be a character or categorical")

        if group_by == color_by:
            raise ValueError("'group_by' and 'color_by' cannot be identical")

    if color_by not in data.columns:
        raise ValueError(f"'{color_by}' is missing from DataFrame")

    if not pd.api.types.is_categorical_dtype(
        data[color_by]
    ) and not pd.api.types.is_string_dtype(data[color_by]):
        raise ValueError("'color_by' must be a character or categorical")

    # Create plot
    if group_by is not None:
        grouped_data = data.groupby([group_by, color_by]).size().reset_index(name="n")
        if flip_axes:
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                p = sns.barplot(
                    x="n", y=group_by, hue=color_by, data=grouped_data, dodge=True
                )
            p.set_xlabel("Count")
            p.set_ylabel(group_by)
        else:
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                p = sns.barplot(
                    x=group_by, y="n", hue=color_by, data=grouped_data, dodge=True
                )
            p.set_ylabel("Count")
            p.set_xlabel(group_by)

        p.set_title(f"Cell counts per {color_by} split by {group_by}")
    else:
        grouped_data = data.groupby(color_by).size().reset_index(name="n")
        if flip_axes:
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                p = sns.barplot(x="n", y=color_by, hue=color_by, data=grouped_data)
            p.set_xlabel("Count")
            p.set_ylabel(color_by)
        else:
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                p = sns.barplot(x=color_by, y="n", hue=color_by, data=grouped_data)
            p.set_ylabel("Count")
            p.set_xlabel(color_by)
        p.set_title(f"Cell counts per {color_by}")

    return plt.gcf(), plt.gca()


def _plot_joint_distribution(data, x, y, show_marginal, **kargs):
    g = sns.JointGrid(data, x=x, y=y)
    g.plot_marginals(sns.kdeplot, fill=True)
    g.ax_marg_x.axis("off")
    g.ax_marg_y.axis("off")
    g.plot_joint(
        sns.scatterplot,
        legend=False,
        data=data,
        hue="density",
        palette=jet_colormap,
        size=0.1,
    )
    if not show_marginal:
        g.ax_marg_x.set_visible(False)
        g.ax_marg_y.set_visible(False)
    return g


def _add_gate_box(
    data,
    gate: pd.Series | pd.DataFrame,
    marker1,
    marker2,
    facet_row=None,
    facet_column=None,
    ax=None,
    **kargs,
):
    if ax is None:
        ax = plt.gca()

    if facet_row is not None and facet_column is not None:
        condition = (data.iloc[0][facet_row], data.iloc[0][facet_column])
    elif facet_row is not None:
        condition = data.iloc[0][facet_row]
    elif facet_column is not None:
        condition = data.iloc[0][facet_column]
    else:
        condition = None

    if isinstance(gate, pd.DataFrame):
        if condition in gate.index:
            gate = gate.loc[condition, :]
        else:
            return
    ax.add_patch(
        Rectangle(
            xy=(gate["xmin"], gate["ymin"]),
            width=gate["xmax"] - gate["xmin"],
            height=gate["ymax"] - gate["ymin"],
            fill=False,
            linestyle="--",
            linewidth=1,
            edgecolor="black",
        )
    )

    # Counting data points inside the gate box
    inside_box = (
        (data[marker1] >= gate["xmin"])
        & (data[marker1] < gate["xmax"])
        & (data[marker2] >= gate["ymin"])
        & (data[marker2] < gate["ymax"])
    )
    inside_percentage = f"{inside_box.mean() * 100:.1f}%"
    ax.text(
        x=gate["xmin"],
        y=gate["ymax"],
        s=inside_percentage,
        verticalalignment="top",
        horizontalalignment="left",
        fontsize=10,
        color="black",
    )
    return


def density_scatter_plot(
    adata: anndata.AnnData,
    marker1: str,
    marker2: str,
    layer: str | None = None,
    facet_row: str | None = None,
    facet_column: str | None = None,
    gate: pd.Series | pd.DataFrame | None = None,
    show_marginal=False,
):
    """Pseudocolor density scatter plot.

    This function creates a scatter plot of abundance data for two markers colored
    based on the density of points:
    Example usage: `density_scatter_plot(pxl.adata, "CD3E", "CD4")`.
    It is also possible to specify one or two variables for faceting the plot by
    columns and rows:
    Example usage: `density_scatter_plot(pxl.adata, "CD3E", "CD4",
                        facet_row = "stimulation", facet_column = "donor")`.
    `facet_row` and `facet_column` should be names of categorical columns in `adata.obs`.
    In addition, a gate can be specified as a Series with xmin, xmax, ymin, and ymax
    to mark a range for components of interest. Alternatively, gate can be specified
    as a DataFrame to allow for different gate ranges for various conditions.
    When both facet_row and facet_column are specified, the condition becomes a
    tuple (facet_row, facet_column), if only one is specified, that parameter
    becomes the condition. The condition permuations are used as the index of
     the gate.
    Example usage:
        gate = pd.DataFrame(columns = ["xmin", "ymin", "xmax", "ymax"])
        gate.loc["Resting"] = [2, 2, 5, 4]
        gate.loc["PHA"] = [1.5, 1.5, 5, 4]
        fig, ax = density_scatter_plot(pixel.adata, "CD3E", "CD4", layer="dsb",
                    facet_column="sample", gate=gate)


    :param adata: Anndata object containing the marker abundance data per
     component.
    :param marker1: The first marker to plot (x-axis).
    :param marker2: The second marker to plot (y-axis).
    :param layer: The anndata layer (e.g. transformation) to use for the marker
     data. Defaults to None.
    :param facet_row: The column to use for faceting the plot by rows.
     Defaults to None.
    :param facet_column: The column to use for faceting the plot by columns.
     Defaults to None.
    :param gate: The gate to use for marking a range of interest. Defaults to
     None.
    :param show_marginal: Whether to show marginal distributions. Defaults to False.
    """
    layer_data = adata.to_df(layer)
    data = layer_data.loc[:, [marker1, marker2]]
    data.loc[:, "density"] = gaussian_kde(data.T)(data.T)
    if facet_column is not None:
        data.loc[:, facet_column] = adata.obs.loc[:, facet_column]

    if facet_row is not None:
        data.loc[:, facet_row] = adata.obs.loc[:, facet_row]

    if facet_column is not None or facet_row is not None:
        if show_marginal:
            warnings.warn("show_marginal is not supported for faceted plots.")
        plot_grid = sns.FacetGrid(data=data, col=facet_column, row=facet_row)
        plot_grid.map_dataframe(
            sns.scatterplot,
            x=marker1,
            y=marker2,
            hue="density",
            palette=jet_colormap,
            size=0.1,
        )
        if gate is not None:
            plot_grid.map_dataframe(
                _add_gate_box,
                gate=gate,
                marker1=marker1,
                marker2=marker2,
                facet_row=facet_row,
                facet_column=facet_column,
            )
        plot_grid.refline(x=0, y=0)
    else:
        plot_grid = _plot_joint_distribution(
            data, x=marker1, y=marker2, show_marginal=show_marginal
        )
        if gate is not None:
            _add_gate_box(
                data, marker1=marker1, marker2=marker2, ax=plot_grid.ax_joint, gate=gate
            )
        plot_grid.ax_joint.axhline(0, color="black", linewidth=1, linestyle="--")
        plot_grid.ax_joint.axvline(0, color="black", linewidth=1, linestyle="--")
    return plt.gcf(), plt.gca()


def abundance_colocalization_plot(
    pixel,
    markers_x: list[str],
    markers_y: list[str],
    layer=None,
    colocalization_column="pearson_z",
):
    """Plot abundance of markers x and y with colocalization as color.

    :param pixel: Pixel object containing the data.
    :param markers_x: List of markers for the x-axis.
    :param markers_y: List of markers for the y-axis.
    :param layer: The anndata layer (e.g. transformation) to use for the marker data.
    :param colocalization_column: The column in the colocalization table to use for
        colocalization values. Defaults to "pearson_z".
    :return: a scatter plot of marker abundance with colocalization as color.
    """
    data = pixel.adata.to_df(layer)
    merged_data = pd.DataFrame()
    for i, mx in enumerate(markers_x):
        for j, my in enumerate(markers_y):
            marker_pair_rows = (
                (pixel.colocalization["marker_1"] == mx)
                & (pixel.colocalization["marker_2"] == my)
            ) | (
                (pixel.colocalization["marker_1"] == my)
                & (pixel.colocalization["marker_2"] == mx)
            )

            coloc_data = pixel.colocalization.loc[marker_pair_rows, :].set_index(
                "component"
            )[colocalization_column]
            data["colocalization"] = coloc_data
            data["colocalization_abs"] = data["colocalization"].abs()
            data["x_abundance"] = data[mx]
            data["y_abundance"] = data[my]
            data["marker_x"] = mx
            data["marker_y"] = my
            data.fillna(0)
            merged_data = pd.concat((merged_data, data), axis=0)
    plot_grid = sns.FacetGrid(data=merged_data, col="marker_x", row="marker_y")
    plot_grid.map_dataframe(
        sns.scatterplot,
        x="x_abundance",
        y="y_abundance",
        hue="colocalization",
        size="colocalization_abs",
        hue_norm=Normalize(
            vmin=merged_data["colocalization"].quantile(0.1),
            vmax=merged_data["colocalization"].quantile(0.9),
            clip=True,
        ),
        size_norm=Normalize(
            vmin=merged_data["colocalization_abs"].quantile(0.1),
            vmax=merged_data["colocalization_abs"].quantile(0.9),
            clip=True,
        ),
    )
    # TODO: See how the legend is determined based on the merged data to be
    # able to access actual marker sizes. Right now the _legend_data includes
    # many points with different sizes and and colors even though both are
    # normalized to the same range. So in the code below, we are only keeping
    # the first 5 points to include the color range and setting all the sizes
    # to a fixed value 5.
    for i in range(1, 6):
        if i >= len(plot_grid._legend_data):
            break
        plot_grid._legend_data[list(plot_grid._legend_data.keys())[i]].set_markersize(5)
    legend_data = {
        i: plot_grid._legend_data[i] for i in list(plot_grid._legend_data.keys())[:6]
    }
    plot_grid.add_legend(legend_data=legend_data)
    return plt.gcf(), plt.gca()



"""Functions for creating plots that are useful with PNA data.

Copyright © 2025 Pixelgen Technologies AB.
"""

from typing import Optional, Tuple

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def molecule_rank_plot(
    data: pd.DataFrame, group_by: Optional[str] = None
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot the number of molecules (n_umi) per component against its n_umi rank.

    Args:
        data (pd.DataFrame): A pandas DataFrame with a column 'n_umi' containing
            antibody counts in components.
        group_by (Optional[str]): A column in the DataFrame to group the plot by.

    Returns:
        Tuple[plt.Figure, plt.Axes]: A plot showing the number of molecules per
        component against its molecule rank used for quality control.

    Raises:
        AssertionError: If the required column(s) are not present in the DataFrame.
        ValueError: If the data types are invalid.

    """
    if "n_umi" not in data.columns:
        raise AssertionError("column 'n_umi' is missing from DataFrame")

    if group_by is not None:
        if group_by not in data.columns:
            raise AssertionError(f"group variable '{group_by}' not found in DataFrame")

        if data[group_by].dtype not in ["object", "category"]:
            raise ValueError(
                f"Invalid class '{data[group_by].dtype}' for column '{group_by}'. "
                f"Expected a string or categorical value"
            )
        else:
            molecule_rank_df = data[[group_by, "n_umi"]].copy()
            molecule_rank_df["rank"] = molecule_rank_df.groupby([group_by])[
                "n_umi"
            ].rank(ascending=False, method="first")
    else:
        molecule_rank_df = data[["n_umi"]].copy()
        molecule_rank_df["rank"] = molecule_rank_df["n_umi"].rank(
            ascending=False, method="first"
        )

    plot_grid = (
        sns.relplot(
            data=molecule_rank_df,
            x="rank",
            y="n_umi",
            hue=group_by,
            aspect=1.6,
        )
        .set(xscale="log", yscale="log")
        .set_xlabels("Component rank (by number of molecules)")
        .set_ylabels("Number of molecules")
    )

    return plot_grid.figure, plot_grid.ax



"""Common click CLI helpers for the Pixelator PNA CLI.

Copyright © 2024 Pixelgen Technologies AB
"""

import functools
import logging
from pathlib import Path

import click

from pixelator.common.utils.units import parse_size


def output_option(func):
    """Wrap a Click entrypoint to add the --output option."""

    @click.option(
        "--output",
        required=True,
        type=click.Path(exists=False),
        help=(
            "The path where the results will be placed (it is created if it does not"
            " exist)"
        ),
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def threads_option(func):
    """Decorate a click command and add the --threads option."""

    @click.option(
        "--threads",
        default=-1,
        required=False,
        show_default=True,
        help="The number of total worker threads available for parallel processing",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def _memory_validator(ctx, param, value):
    if value is None:
        return None
    try:
        return int(parse_size(value))
    except ValueError as exc:
        raise click.BadParameter(
            "--memory option must be a positive integer, optionally with a unit suffix [K, M, G]"
        )


def memory_option(func):
    """Decorate a click command and add the --memory option."""

    @click.option(
        "--memory",
        default=None,
        required=False,
        callback=_memory_validator,
        show_default=True,
        help="The maximum amount of memory available for processing",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def design_option(func):
    """Decorate a click command and add the --design option."""
    from pixelator.pna.config import pna_config

    # TODO: Support assay aliases here as well
    #   implement pna_config.list_assay_names()
    assay_options = list(pna_config.assays.keys())

    @click.option(
        "--design",
        required=True,
        default=None,
        type=click.Choice(assay_options),
        help="The design to load from the configuration file",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def validate_panel(ctx, param, value):
    """Validate the panel commandline option.

    :param ctx: The click context
    :param param: The click parameter
    :param value: The click value
    :returns: The validated value
    """
    try:
        if Path(value).exists():
            return value
    except Exception:
        pass
    else:
        from pixelator.pna.config import pna_config

        panel_options = list(pna_config.list_panel_names(include_aliases=True))

        if value not in panel_options:
            raise click.UsageError(
                f"Panel {value} is not a file and is not an id of a supported panel."
            )

    return value


def panel_option(func):
    """Decorate a click command and add the --design option."""
    from pixelator.pna.config import pna_config

    # TODO: Support assay aliases here as well
    #   implement pna_config.list_assay_names()
    panel_options = list(pna_config.list_panel_names())

    @click.option(
        "--panel",
        required=True,
        default=None,
        type=click.UNPROCESSED,
        callback=validate_panel,
        help="The name of a panel to load from the supported panels. Optionally, provide a path to a custom panel file.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


logger = logging.getLogger("pixelator.pna.cli")



"""Collect statistics for amplicon.

Copyright © 2023 Pixelgen Technologies AB.
"""

import collections
import dataclasses
from functools import cache
from typing import Any, Union

import numba as nb
import numpy as np

from pixelator.common.config import RegionType, get_position_in_parent
from pixelator.mpx.config import config


@nb.njit
def _count_elem_in_array_where_greater_than(arr, value):
    result = 0
    for x in arr:
        if x > value:
            result += 1
    return result


@nb.njit
def _count_elem_in_array_where_greater_or_equal_than(arr, value):
    result = 0
    for x in arr:
        if x >= value:
            result += 1
    return result


@dataclasses.dataclass(frozen=True, slots=True)
class SequenceQualityStats:
    """Container for sequence quality statistics."""

    fraction_q30_upia: float
    fraction_q30_upib: float
    fraction_q30_umi: float
    fraction_q30_pbs1: float
    fraction_q30_pbs2: float
    fraction_q30_bc: float
    fraction_q30: float

    def asdict(self) -> dict[str, Any]:
        """Return a dictionary representation of this instance."""
        return {k: v for k, v in dataclasses.asdict(self).items()}


class SequenceQualityStatsCollector:
    """Accumulate read quality statistics for a given design."""

    def __init__(self, design_name: str):
        """Accumulate read quality statistics for a given design.

        :param design_name: The name of the design of the reads for which to statistics.
        """
        design = config.get_assay(design_name)

        if design is None:
            raise ValueError(f"Unknown design: {design_name}")

        self.design = design
        self._counter: collections.Counter[str] = collections.Counter()
        self._positions = {}

        self._amplicon_len = len(design)

        amplicon = self.design.get_region_by_id("amplicon")
        if amplicon is None:
            raise ValueError("Assay does not contain an amplicon region")

        for leave in amplicon.get_leaves():
            self._positions[leave.region_id] = get_position_in_parent(
                self.design, leave.region_id
            )

        # Verify all the right regions are present
        # Check for upi, pbs and bc regions
        for region_id in ("upi-a", "upi-b", "pbs-1", "pbs-2", "bc"):
            if region_id not in self._positions:
                raise ValueError(f"Assay does not contain region {region_id}")

        # Check that at least one umi region is present
        umi_regions = self.design.get_regions_by_type(RegionType.UMI)
        if not umi_regions:
            raise ValueError("Assay does not contain a UMI region")

    @cache
    def get_position(self, region_id: str) -> tuple[int, int]:
        """Return the positions for a region.

        :param region_id: id of the region
        :returns: a tuple with start and end positions
        :raise ValueError: An unknown region id was given
        """
        r = self._positions.get(region_id)
        if r is None:
            raise ValueError(f"Unknown region: {region_id}")
        return r

    @staticmethod
    def _read_stats(quali: np.ndarray) -> tuple[int, int]:
        bases_in_read = _count_elem_in_array_where_greater_than(quali, 2)
        q30_bases_in_read = _count_elem_in_array_where_greater_or_equal_than(quali, 30)
        return bases_in_read, q30_bases_in_read

    def _umi_stats(self, quali: np.ndarray) -> tuple[int, int]:
        # Q30 Bases in UMI
        umi_regions = self.design.get_regions_by_type(RegionType.UMI)
        umi_positions = [self.get_position(r.region_id) for r in umi_regions]

        p = np.zeros(self._amplicon_len, dtype=np.uint8)
        for pos in umi_positions:
            p[slice(*pos)] = 1

        q = p * quali

        bases_in_umi = _count_elem_in_array_where_greater_than(q, 2)
        q30_bases_in_umi = _count_elem_in_array_where_greater_or_equal_than(q, 30)

        return bases_in_umi, q30_bases_in_umi

    def _get_stats_from_position(self, quali: np.ndarray, pos: str) -> tuple[int, int]:
        upia_pos = self.get_position(pos)
        slice_obj = slice(*upia_pos)
        quali_subset = quali[slice_obj]
        bases = _count_elem_in_array_where_greater_than(quali_subset, 2)
        q30 = _count_elem_in_array_where_greater_or_equal_than(quali_subset, 30)
        return bases, q30

    def _upia_stats(self, quali: np.ndarray) -> tuple[int, int]:
        return self._get_stats_from_position(quali, "upi-a")

    def _upib_stats(self, quali: np.ndarray) -> tuple[int, int]:
        return self._get_stats_from_position(quali, "upi-b")

    def _pbs1_stats(self, quali: np.ndarray) -> tuple[int, int]:
        return self._get_stats_from_position(quali, "pbs-1")

    def _pbs2_stats(self, quali: np.ndarray) -> tuple[int, int]:
        return self._get_stats_from_position(quali, "pbs-2")

    def _bc_stats(self, quali: np.ndarray) -> tuple[int, int]:
        return self._get_stats_from_position(quali, "bc")

    @property
    def read_count(self) -> int:
        """Return the number of reads processed."""
        return self._counter["read_count"]

    @property
    def stats(self) -> SequenceQualityStats:
        """Return the accumulated statistics as a SequenceQualityStats object."""
        fraction_q30_upia = (
            self._counter["q30_bases_in_upia"] / self._counter["bases_in_upia"]
        )
        fraction_q30_upib = (
            self._counter["q30_bases_in_upib"] / self._counter["bases_in_upib"]
        )
        fraction_q30_umi = (
            self._counter["q30_bases_in_umi"] / self._counter["bases_in_umi"]
        )
        fraction_q30_pbs1 = (
            self._counter["q30_bases_in_pbs1"] / self._counter["bases_in_pbs1"]
        )
        fraction_q30_pbs2 = (
            self._counter["q30_bases_in_pbs2"] / self._counter["bases_in_pbs2"]
        )
        fraction_q30_bc = (
            self._counter["q30_bases_in_bc"] / self._counter["bases_in_bc"]
        )
        fraction_q30 = (
            self._counter["q30_bases_in_read"] / self._counter["bases_in_read"]
        )

        return SequenceQualityStats(
            fraction_q30_upia=fraction_q30_upia,
            fraction_q30_upib=fraction_q30_upib,
            fraction_q30_umi=fraction_q30_umi,
            fraction_q30_pbs1=fraction_q30_pbs1,
            fraction_q30_pbs2=fraction_q30_pbs2,
            fraction_q30_bc=fraction_q30_bc,
            fraction_q30=fraction_q30,
        )

    def update(self, qualities: Union[str, np.ndarray]) -> None:
        """Update the statistics with the given read qualities."""
        # Use numpy for vectorized operations
        # Reinterpret cast to integers (same as ord)
        if isinstance(qualities, str):
            quali = np.frombuffer(qualities.encode(), dtype=np.uint8) - 33
        else:
            quali = qualities

        bases_in_read, q30_bases_in_read = self._read_stats(quali)
        bases_in_umi, q30_bases_in_umi = self._umi_stats(quali)
        bases_in_upia, q30_bases_in_upia = self._upia_stats(quali)
        bases_in_upib, q30_bases_in_upib = self._upib_stats(quali)
        bases_in_pbs1, q30_bases_in_pbs1 = self._pbs1_stats(quali)
        bases_in_pbs2, q30_bases_in_pbs2 = self._pbs2_stats(quali)
        bases_in_bc, q30_bases_in_bc = self._bc_stats(quali)

        self._counter.update(
            read_count=1,
            bases_in_read=bases_in_read,
            q30_bases_in_read=q30_bases_in_read,
            bases_in_umi=bases_in_umi,
            q30_bases_in_umi=q30_bases_in_umi,
            bases_in_upia=bases_in_upia,
            q30_bases_in_upia=q30_bases_in_upia,
            bases_in_upib=bases_in_upib,
            q30_bases_in_upib=q30_bases_in_upib,
            bases_in_pbs1=bases_in_pbs1,
            q30_bases_in_pbs1=q30_bases_in_pbs1,
            bases_in_pbs2=bases_in_pbs2,
            q30_bases_in_pbs2=q30_bases_in_pbs2,
            bases_in_bc=bases_in_bc,
            q30_bases_in_bc=q30_bases_in_bc,
        )



"""Module with functions that relate to marking functions in pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

import warnings
from functools import wraps


def experimental(f):
    """Annotation for experimental pixelator functions."""

    @wraps(f)
    def wrapper(*args, **kwds):
        warnings.warn(
            f"The function `{f.__name__}` is experimental, it might be removed or "
            "the API might change without notice.",
            stacklevel=2,
        )
        return f(*args, **kwds)

    return wrapper



"""Keep constants used on graph operations in pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

MIN_PIXELS_TO_REFINE = 100
LEIDEN_RESOLUTION = 0.01
RELATIVE_ANNOTATE_RESOLUTION = (
    0.5  # A lower resolution is used for annotation of potential doublets
)



"""Functions and classes related to reporting for the collapse stage.

Copyright © 2024 Pixelgen Technologies AB
"""

import pydantic

from pixelator.pna.collapse.paired.statistics import (
    CollapseInputFile,
    MarkerLinkGroupStats,
)
from pixelator.pna.report.models.base import SampleReport


class CollapseSampleReport(SampleReport):
    """Model for report data returned by the collapse stage."""

    report_type: str = "collapse"
    input_reads: int = pydantic.Field(
        ..., description="The number of input reads processed."
    )

    input_molecules: int = pydantic.Field(
        ...,
        description="The number of unique molecules (unique input reads) processed.",
    )

    @pydantic.computed_field(  # type: ignore
        description="The number of output reads.", return_type=int
    )
    @property
    def output_reads(self) -> int:
        """The total number of error-corrected output reads.

        This is an alias for the input reads as no reads are removed during the collapse step.
        """
        return self.input_reads

    output_molecules: int = pydantic.Field(
        ..., description="The total number of error-corrected molecules detected."
    )

    unique_marker_links: int = pydantic.Field(
        ...,
        description="The total number of unique marker links (edges) in the output molecules.",
    )

    processed_files: list[CollapseInputFile] = pydantic.Field(
        ..., description="The files processed during the collapse step."
    )
    markers: list[MarkerLinkGroupStats] = pydantic.Field(
        ..., description="The statistics for each marker pair."
    )



"""Copyright © 2023 Pixelgen Technologies AB."""

from .cli_info import CLIInvocationInfo, CommandInfo, CommandOption
from .reporting import PixelatorPNAReporting
from .workdir import PixelatorPNAWorkdir, SingleCellPNAStage, WorkdirOutputNotFound

__all__ = [
    "CLIInvocationInfo",
    "CommandInfo",
    "CommandOption",
    "PixelatorPNAWorkdir",
    "PixelatorPNAReporting",
    "SingleCellPNAStage",
    "WorkdirOutputNotFound",
]



"""Build an amplicon from paired-end reads.

Copyright © 2024 Pixelgen Technologies AB.
"""

import enum
import typing
from dataclasses import dataclass
from typing import Any, Optional, cast

import numpy as np
from cutadapt._align import Aligner
from cutadapt.adapters import Where
from cutadapt.info import ModificationInfo
from cutadapt.statistics import ReadLengthStatistics
from cutadapt.steps import HasFilterStatistics
from dnaio import SequenceRecord

from pixelator.pna.config import SequenceType
from pixelator.pna.config.assay import PNAAssay, Region, get_position_in_parent
from pixelator.pna.read_processing.modifiers import CombiningModifier
from pixelator.pna.read_processing.statistics import HasCustomStatistics

#: Translation table for reverse complementing DNA sequences (strings)
_TRTABLE = str.maketrans("GTACN", "CATGN")

#: Translation table for reverse complementing DNA sequences (bytes)
_TRTABLE_BYTES = bytes.maketrans(b"GTACN", b"CATGN")

BytesOrString = typing.TypeVar("BytesOrString", str, bytes)


def reverse_complement(seq: BytesOrString) -> BytesOrString:
    """Compute the reverse complement of a DNA seq.

    :param seq: the DNA sequence
    :return: the reverse complement of the input sequence
    :rtype: Bytes or Str depending on the input
    """
    if isinstance(seq, bytes):
        return seq.translate(_TRTABLE_BYTES)[::-1]

    return seq.translate(_TRTABLE)[::-1]


class AmpliconBuilderStatistics:
    """Helper class for AmpliconBuilder to keep track of pipeline statistics."""

    def __init__(self):
        """Initialize the statistics."""
        self.passed_reads = 0
        self.failed_reads = 0

        self.passed_missing_lbs1_anchor = 0
        self.passed_partial_uei_reads = 0
        self.failed_partial_upi1_umi1_reads = 0
        self.failed_partial_upi2_umi2_reads = 0

        self.passed_missing_uei_reads = 0
        self.failed_missing_upi1_umi1_reads = 0
        self.failed_missing_upi2_umi2_reads = 0

    def __iadd__(self, other):
        """Merge statistics from another object into this one."""
        if not isinstance(other, self.__class__):
            raise ValueError("Cannot compare")

        self.passed_reads += other.passed_reads
        self.failed_reads += other.failed_reads

        self.passed_missing_lbs1_anchor += other.passed_missing_lbs1_anchor
        self.passed_partial_uei_reads += other.passed_partial_uei_reads
        self.failed_partial_upi1_umi1_reads += other.failed_partial_upi1_umi1_reads
        self.failed_partial_upi2_umi2_reads += other.failed_partial_upi2_umi2_reads

        self.passed_missing_uei_reads += other.passed_missing_uei_reads
        self.failed_missing_upi1_umi1_reads += other.failed_missing_upi1_umi1_reads
        self.failed_missing_upi2_umi2_reads += other.failed_missing_upi2_umi2_reads

        return self

    def collect(self) -> dict[str, int]:
        """Return a dictionary with statistics."""
        return {
            "passed_reads": self.passed_reads,
            "passed_missing_eui_reads": self.passed_missing_uei_reads,
            "failed_reads": self.failed_reads,
            "passed_partial_uei_reads": self.passed_partial_uei_reads,
            "passed_missing_lbs1_anchor": self.passed_missing_lbs1_anchor,
            "failed_partial_upi1_umi1_reads": self.failed_partial_upi1_umi1_reads,
            "failed_partial_upi2_umi2_reads": self.failed_partial_upi2_umi2_reads,
            "failed_missing_upi1_umi1_reads": self.failed_missing_upi1_umi1_reads,
            "failed_missing_upi2_umi2_reads": self.failed_missing_upi2_umi2_reads,
        }

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"AmpliconBuilderStatistics(passed_reads={self.passed_reads}, failed_reads={self.failed_reads})"


class AmpliconCombiner:
    """Helper class for `AmpliconBuilder` to assemble an amplicon from different regions."""

    _REQUIRED_REGIONS = {
        "amplicon",
        "lbs-1",
        "lbs-2",
        "pid-1",
        "pid-2",
        "umi-1",
        "umi-2",
        "uei",
    }

    def __init__(self, assay: PNAAssay):
        """Initialize the AmpliconCombiner.

        :param assay: the assay design
        """
        self.assay = assay
        self._template = (
            cast(Region, assay.get_region_by_id("amplicon"))
            .get_sequence()
            .encode("ascii")
        )
        self._qualities_template = bytes(b"!" * len(self._template))
        self._uei_template = bytes(b"N" * assay.get_region_by_id("uei").min_len)
        self._uei_qualities_template = bytes(b"!" * len(self._uei_template))

        lbs1_len = cast(Region, assay.get_region_by_id("lbs-1")).min_len
        pid1_len = cast(Region, assay.get_region_by_id("pid-1")).min_len
        umi1_len = cast(Region, assay.get_region_by_id("umi-1")).min_len
        uei_len = cast(Region, assay.get_region_by_id("uei")).min_len
        lbs2_len = cast(Region, assay.get_region_by_id("lbs-2")).min_len
        pid2_len = cast(Region, assay.get_region_by_id("pid-2")).min_len
        umi2_len = cast(Region, assay.get_region_by_id("umi-2")).min_len

        self._pid1_umi1_region_slice = slice(0, pid1_len + umi1_len)
        self._uei_region_slice = slice(
            pid1_len + umi1_len + lbs1_len, pid1_len + umi1_len + lbs1_len + uei_len
        )
        self._pid2_umi2_region_slice = slice(
            pid1_len + umi1_len + lbs1_len + uei_len + lbs2_len,
            pid1_len + umi1_len + lbs1_len + uei_len + lbs2_len + pid2_len + umi2_len,
        )
        self._lbs1_region_slice = slice(*get_position_in_parent(assay, "lbs-1"))
        self._lbs2_region_slice = slice(*get_position_in_parent(assay, "lbs-2"))

    def build_sequence(
        self, pid1_umi1_region: bytes, pid2_umi2_region: bytes, uei_region: bytes | None
    ) -> bytes:
        """Build the amplicon sequence from the different regions.

        :param pid1_umi1_region: the PBS-1 and UMI-1 region
        :param uei_region: the UEI region
        :param pid2_umi2_region: the PBS-2 and UMI-2 region
        """
        # We are using bytearray here since strings are immutable
        s = bytearray(self._template)
        s[self._pid1_umi1_region_slice] = pid1_umi1_region
        s[self._pid2_umi2_region_slice] = pid2_umi2_region

        if uei_region:
            uei_len = len(self._uei_template)
            insert_len = len(uei_region)
            if insert_len >= uei_len:
                s[self._uei_region_slice] = uei_region[:uei_len]
            else:
                s[self._uei_region_slice] = self._uei_template
                s[
                    self._uei_region_slice.start : self._uei_region_slice.start
                    + insert_len
                ] = uei_region
        else:
            s[self._uei_region_slice] = self._uei_template

        # LBS-1 and LBS-2 are already included in the template
        return s

    def build_qualities(
        self,
        pid1_umi1_region: bytes,
        pid2_umi2_region: bytes,
        uei_region: bytes | None = None,
        lbs1_region: bytes | None = None,
        lbs2_region: bytes | None = None,
    ) -> bytes:
        """Build the amplicon qualities from the different regions.

        All regions are assumed to have the correct length.

        :param pid1_umi1_region: the PBS-1 and UMI-1 region qualities
        :param uei_region: the UEI region qualities
        :param pid2_umi2_region: the PBS-2 and UMI-2 region qualities
        :param lbs1_region: the LBS-1 region qualities
        :param lbs2_region: the LBS-2 region qualities
        """
        # We are using bytearray here since strings are immutable
        s = bytearray(self._qualities_template)
        s[self._pid1_umi1_region_slice] = pid1_umi1_region
        s[self._pid2_umi2_region_slice] = pid2_umi2_region

        if uei_region:
            uei_len = len(self._uei_template)
            insert_len = len(uei_region)
            if insert_len >= uei_len:
                s[self._uei_region_slice] = uei_region[:uei_len]
            else:
                s[self._uei_region_slice] = self._uei_qualities_template
                s[
                    self._uei_region_slice.start : self._uei_region_slice.start
                    + insert_len
                ] = uei_region
        else:
            s[self._uei_region_slice] = self._uei_qualities_template

        if lbs1_region is not None:
            assert len(lbs1_region) == (
                self._lbs1_region_slice.stop - self._lbs1_region_slice.start
            )
            s[self._lbs1_region_slice] = lbs1_region
        if lbs2_region is not None:
            assert len(lbs2_region) == (
                self._lbs2_region_slice.stop - self._lbs2_region_slice.start
            )
            s[self._lbs2_region_slice] = lbs2_region

        return s


class AmpliconRegionSlices:
    """Helper class to store slices for the different regions of a PNA amplicon."""

    __slots__ = ("pid1_umi1", "pid2_umi2", "uei", "lbs1", "lbs2")

    def __init__(self):
        """Initialize the AmpliconRegionSlices instance."""
        self.pid1_umi1 = None
        self.pid2_umi2 = None
        self.uei = None
        self.lbs1 = None
        self.lbs2 = None

    def __getitem__(self, item):
        """Provide tuple like access to the slices."""
        if isinstance(item, int):
            key = self.__slots__[item]
            return getattr(self, key)

    def __repr__(self):
        """Return a string representation of the object."""
        return (
            f"AmpliconRegionSlices(pid1_umi1={self.pid1_umi1}, pid2_umi2={self.pid2_umi2}, "
            f"uei={self.uei}, lbs1={self.lbs1}, lbs2={self.lbs2})"
        )


class AmpliconBuilderFailureReason(enum.Enum):
    """Reasons for failing to build an amplicon."""

    MISSING_PID1_UMI1 = "missing_pid1_umi1"
    MISSING_PID2_UMI2 = "missing_pid2_umi2"
    MISSING_UEI = "missing_uei"
    PARTIAL_PID1_UMI1 = "partial_pid1_umi1"
    PARTIAL_PID2_UMI2 = "partial_pid2_umi2"
    PARTIAL_UEI = "partial_uei"
    NO_CONSENSUS = "no_consensus"


class AmpliconBuilder(CombiningModifier, HasFilterStatistics, HasCustomStatistics):
    """Construct an amplicon from a pair of reads.

    :param assay: the assay design
    :param mismatches: the maximum number of mismatches allowed when aligning the LBS sequences
    :param writer: a writer to save failed reads to
    """

    def __init__(self, assay: PNAAssay, mismatches: int | float = 0.2, writer=None):
        """Initialize the AmpliconBuilder.

        :param assay: the assay design
        :param mismatches: the maximum number of mismatches allowed when aligning the LBS sequences.
            If a float, it is interpreted as a fraction of the template length. Otherwise, it is
            the maximum number of mismatches allowed.
        :param writer: a writer to save failed reads to.
        """
        self._assay = assay
        self._writer = writer
        self._region_combiner = AmpliconCombiner(assay)

        amplicon_len = assay.get_region_by_id("amplicon").min_len
        lbs1_ref = assay.get_region_by_id("lbs-1").get_sequence()
        lbs2_ref = assay.get_region_by_id("lbs-2").get_sequence()
        lbs1_ref_rc = reverse_complement(lbs1_ref)
        lbs2_ref_rc = reverse_complement(lbs2_ref)

        min_overlap = 4
        indel_cost = 2

        self._lbs1_aligner = Aligner(
            reference=lbs1_ref,
            max_error_rate=mismatches,
            wildcard_ref=False,
            wildcard_query=True,
            indel_cost=indel_cost,
            min_overlap=min_overlap,
            flags=Where.BACK,
        )

        self._lbs2_aligner = Aligner(
            reference=lbs2_ref,
            max_error_rate=mismatches,
            wildcard_ref=False,
            wildcard_query=True,
            indel_cost=indel_cost,
            min_overlap=min_overlap,
            flags=Where.BACK,
        )

        self._lbs1_rc_aligner = Aligner(
            reference=lbs1_ref_rc,
            max_error_rate=mismatches,
            wildcard_ref=False,
            wildcard_query=True,
            indel_cost=indel_cost,
            min_overlap=min_overlap,
            flags=Where.BACK,
        )

        self._lbs2_rc_aligner = Aligner(
            reference=lbs2_ref_rc,
            max_error_rate=mismatches,
            wildcard_ref=False,
            wildcard_query=True,
            indel_cost=indel_cost,
            min_overlap=min_overlap,
            flags=Where.FRONT,
        )

        # Some helper variables
        # We define these here to save a bit of time by not regenerating them for each read
        self._pid_1_umi_1_region_len = (
            assay.get_region_by_id("pid-1").min_len
            + assay.get_region_by_id("umi-1").min_len
        )
        self._pid_2_umi_2_region_len = (
            assay.get_region_by_id("pid-2").min_len
            + assay.get_region_by_id("umi-2").min_len
        )
        self._lbs1_seq = assay.get_region_by_id("lbs-1").get_sequence().encode("ascii")
        self._lbs2_seq = assay.get_region_by_id("lbs-2").get_sequence().encode("ascii")
        self._lbs1_qual = bytes(b"!" * len(self._lbs1_seq))
        self._lbs2_qual = bytes(b"!" * len(self._lbs2_seq))

        self._lbs2_start_pos = get_position_in_parent(assay, "lbs-2")[0]

        self._uei_region_len = assay.get_region_by_id("uei").min_len
        self._uei_region_end_pos = get_position_in_parent(assay, "uei")[1]
        self._uei_region_end_pos_rev = amplicon_len - self._uei_region_end_pos

        # If mismatches is a float, interpret it as a fraction of the template length
        # otherwise it is the maximum edit distance

        self._stats = ReadLengthStatistics()
        self._custom_stats = AmpliconBuilderStatistics()

        self._fixed_regions = self._get_fixed_sites()
        self._fixed_bases_len = sum([r.min_len for r, _ in self._fixed_regions])

    def _get_fixed_sites(self) -> list[tuple[Region, tuple[int, int]]]:
        fixed_regions: list[tuple[Region, tuple[int, int]]] = []
        for r in self._assay.get_regions_by_sequence_type(SequenceType.FIXED):
            pos = get_position_in_parent(self._assay, r.region_id)
            fixed_regions.append((r, pos))

        return fixed_regions

    def descriptive_identifier(self) -> str:
        """Return a descriptive identifier for reporting.

        See `HasFilterStatistics`.
        """
        return "invalid_amplicon"

    def get_statistics_name(self):
        """Return a descriptive identifier for detailed statistics.

        See `HasCustomStatistics`.
        """
        return "amplicon"

    def get_statistics(self) -> AmpliconBuilderStatistics:
        """Return the statistics for the amplicon builder.

        See `HasCustomStatistics`.
        """
        return self._custom_stats

    def filtered(self) -> int:
        """Return the number of reads that have been filtered by this stage."""
        return self._custom_stats.failed_reads

    def _scan_forward_read(self, read: SequenceRecord) -> AmpliconRegionSlices:
        """Scan the forward read for the different regions of the amplicon.

        :param read: the forward read
        :return: the slices for the different regions of the amplicon in the read
        """
        region_slices = AmpliconRegionSlices()

        lbs1_alm = self._lbs1_aligner.locate(read.sequence)
        lbs2_alm = None

        if lbs1_alm:
            lbs1_start_pos, lbs1_end_pos = lbs1_alm[2], lbs1_alm[3]
            region_slices.lbs1 = slice(lbs1_start_pos, lbs1_end_pos)

            if lbs1_start_pos >= self._pid_1_umi_1_region_len:
                region_slices.pid1_umi1 = slice(
                    lbs1_start_pos - self._pid_1_umi_1_region_len, lbs1_start_pos
                )
        elif len(read) >= self._pid_1_umi_1_region_len:
            # No LBS-1 found, we will assume the read is properly anchored and extract the PID-1, UMI-1 region,
            # simply from the expected position.
            region_slices.pid1_umi1 = slice(0, self._pid_1_umi_1_region_len)
            lbs1_end_pos = self._pid_1_umi_1_region_len + len(self._lbs1_seq)

        # if the read is longer than the end of the uei region and thus contains the (partial) lbs-2 region
        if len(read) > self._uei_region_end_pos:
            lbs2_alm = self._lbs2_aligner.locate(read.sequence)
            if lbs2_alm:
                # [lbs1 end pos, lbs2 start pos)
                region_slices.uei = slice(lbs1_end_pos, lbs2_alm[2])

        # Check if we have the full pbs-2, umi-2 region after lbs-2 alignment
        if lbs2_alm:
            lbs2_start_pos, lbs2_end_pos = lbs2_alm[2], lbs2_alm[3]
            region_slices.lbs2 = slice(lbs2_start_pos, lbs2_end_pos)

            # TODO: Allow partial regions here when the full LBS-2 is matched ?
            if (lbs2_end_pos + self._pid_2_umi_2_region_len) <= len(read):
                region_slices.pid2_umi2 = slice(
                    lbs2_end_pos, lbs2_end_pos + self._pid_2_umi_2_region_len
                )

        return region_slices

    def _scan_reverse_read(self, read: SequenceRecord) -> AmpliconRegionSlices:
        """Scan the reverse read for the different regions of the amplicon.

        :param read: the reverse read
        :return: the slices for the different regions of the amplicon in the read
        """
        region_slices = AmpliconRegionSlices()

        lbs2_alm = self._lbs2_rc_aligner.locate(read.sequence)
        lbs1_alm = None

        if lbs2_alm:
            lbs_2_start_pos, lbs_2_end_pos = lbs2_alm[2], lbs2_alm[3]
            region_slices.lbs2 = slice(lbs_2_start_pos, lbs_2_end_pos)
            if lbs_2_start_pos >= self._pid_2_umi_2_region_len:
                region_slices.pid2_umi2 = slice(
                    lbs_2_start_pos - self._pid_2_umi_2_region_len, lbs_2_start_pos
                )

            # if the read is longer than the end of the uei region and thus contains the (partial) lbs-2 region
            if len(read) > self._uei_region_end_pos_rev:
                lbs1_alm = self._lbs1_rc_aligner.locate(read.sequence)

                if lbs1_alm:
                    # [lbs2 end pos, lbs1 start pos)
                    region_slices.uei = slice(lbs2_alm[3], lbs1_alm[2])
        elif len(read) > self._pid_2_umi_2_region_len:
            # No LBS-2 found, we will assume the read is properly anchored and extract the PID-2, UMI-2 region,
            # simply from the expected position.
            region_slices.pid2_umi2 = slice(0, self._pid_2_umi_2_region_len)

        # Check if we have the full pbs-1, umi-1 region after lbs-2 alignment
        if lbs1_alm:
            lbs1_start_pos, lbs1_end_pos = lbs1_alm[2], lbs1_alm[3]
            region_slices.lbs1 = slice(lbs1_start_pos, lbs1_end_pos)

            # TODO: Allow partial regions here when the full LBS-1 is matched?
            if (lbs1_end_pos + self._pid_1_umi_1_region_len) <= len(read):
                region_slices.pid1_umi1 = slice(
                    lbs1_end_pos, lbs1_end_pos + self._pid_1_umi_1_region_len
                )

        return region_slices

    @staticmethod
    def _get_region_sequence(
        read, region_slice, is_reversed=False, as_bytearray=False
    ) -> tuple[bytes | None, bytes | None]:
        """Get the sequence of a region from a read.

        :param read: the read to extract the region from
        :param region_slice: the slice for the region
        :param is_reversed: whether the region is reversed (e.g. LBS-2)
        :param return_ascii: whether to return the sequence as ASCII bytes
        :return: the sequence  and quality of the region as a tuple, or None if the slice is empty
        """
        if not region_slice:
            return None, None

        if is_reversed:
            seq = reverse_complement(read.sequence[region_slice])
            qual = (read.qualities[region_slice])[::-1]
        else:
            seq = read.sequence[region_slice]
            qual = read.qualities[region_slice]

        if as_bytearray:
            return bytearray(seq.encode("ascii")), bytearray(qual.encode("ascii"))
        else:
            return seq.encode("ascii"), qual.encode("ascii")

    @staticmethod
    def _consensus_seq(
        read1: SequenceRecord, read2: SequenceRecord, region1_slice, region2_slice
    ):
        """Combine slices from forward and reverse reads into a single sequence.

        If one of the slices is None, the other slice is returned.
        If two slices are present they are combined based on the quality scores.

        :param read1: the forward read
        :param read2: the reverse read
        :param region1_slice: the slice from the forward read
        :param region2_slice: the slice from the reverse read
        """
        if region1_slice is None and region2_slice is None:
            return None, None

        if not region1_slice or not region2_slice:
            if region1_slice:
                s, q = AmpliconBuilder._get_region_sequence(
                    read1,
                    region1_slice,
                    is_reversed=False,
                )
            else:
                s, q = AmpliconBuilder._get_region_sequence(
                    read2,
                    region2_slice,
                    is_reversed=True,
                )

            return s, q

        else:
            # TODO: Consensus of non LBS regions with partial lengths?
            #   We can inject both r1 and r2 into an all N template with zero qualities

            s1, q1 = AmpliconBuilder._get_region_sequence(
                read1, region1_slice, is_reversed=False, as_bytearray=True
            )
            s2, q2 = AmpliconBuilder._get_region_sequence(
                read2, region2_slice, is_reversed=True, as_bytearray=True
            )

            assert len(s1) == len(s2)  # type: ignore
            assert len(q1) == len(q2)  # type: ignore

            s_np = np.select([q1 > q2, q1 <= q2], [s1, s2])  # type: ignore
            q_np = np.max([q1, q2], axis=0)  # type: ignore

            return s_np.tobytes(), q_np.tobytes()

    def _consensus_qual_lbs1(
        self,
        read1: SequenceRecord | None,
        read2: SequenceRecord | None,
        region1_slice,
        region2_slice,
    ):
        """Combine the quality scores of the LBS-1 region from forward and reverse reads.

        :param read1: the forward read
        :param read2: the reverse read
        :param region1_slice: the slice from the forward read
        :param region2_slice: the slice from the reverse read
        """
        return self._consensus_qual_lbs(
            read1, read2, region1_slice, region2_slice, self._lbs1_qual
        )

    def _consensus_qual_lbs2(
        self,
        read1: SequenceRecord | None,
        read2: SequenceRecord | None,
        region1_slice,
        region2_slice,
    ):
        """Combine the quality scores of the LBS-2 region from forward and reverse reads.

        :param read1: the forward read
        :param read2: the reverse read
        :param region1_slice: the slice from the forward read
        :param region2_slice: the slice from the reverse read
        """
        return self._consensus_qual_lbs(
            read1, read2, region1_slice, region2_slice, self._lbs2_qual
        )

    @staticmethod
    def _consensus_qual_lbs(
        read1: SequenceRecord | None,
        read2: SequenceRecord | None,
        region1_slice,
        region2_slice,
        template_qual,
    ) -> bytes:
        """Combine the quality scores of an LBS region from forward and reverse reads.

        :param read1: the forward read
        :param read2: the reverse read
        :param region1_slice: the slice from the forward read
        :param region2_slice: the slice from the reverse read
        :param template_qual: the quality template for the region
        """
        # We cap the length of the quality string to the length of the template
        # Due to indels the quality string can be longer than the template
        # The quality string is thus not 100% accurate in the presence of indels
        # Since we only use this for Q30 statistics, this is not a big issue
        # For fully accurate quality scores we would need to have the full alignment
        # instead of just the matched region and this has a very high overhead.
        if not region1_slice and not region2_slice:
            return template_qual

        if not region1_slice or not region2_slice:
            tq = bytearray(template_qual)

            if region1_slice:
                q = read1.qualities[region1_slice].encode("ascii")[: len(tq)]  # type: ignore
                tq[: len(q)] = q

            else:
                q = (read2.qualities[region2_slice])[::-1].encode("ascii")[: len(tq)]  # type: ignore
                tq[-len(q) :] = q

            return bytes(tq)

        else:
            tq1 = bytearray(template_qual)
            tq2 = bytearray(template_qual)

            q1 = read1.qualities[region1_slice].encode("ascii")[: len(template_qual)]  # type: ignore
            tq1[: len(q1)] = q1

            q2 = (read2.qualities[region2_slice])[::-1].encode("ascii")[  # type: ignore
                : len(template_qual)
            ]
            tq2[-len(q2) :] = q2

            assert len(tq1) == len(tq2)
            tq_numpy = np.max([tq1, tq2], axis=0)

            return tq_numpy.tobytes()

    def _check_regions(
        self, pid1_umi1_region: str, uei_region: str, pid2_umi2_region: str
    ) -> AmpliconBuilderFailureReason | None:
        """Check for valid regions in the read pair.

        Return True if all regions are valid, False otherwise.
        Failure counters are incremented in the statistics object.

        :param pid1_umi1_region: the PID-1 and UMI-1 region
        :param uei_region: the UEI region
        :param pid2_umi2_region: the PID-2 and UMI-2 region
        :return: None if all regions are valid, a failure reason otherwise
        """
        if not pid1_umi1_region:
            self._custom_stats.failed_missing_upi1_umi1_reads += 1
            return AmpliconBuilderFailureReason.MISSING_PID1_UMI1

        if not pid2_umi2_region:
            self._custom_stats.failed_missing_upi2_umi2_reads += 1
            return AmpliconBuilderFailureReason.MISSING_PID2_UMI2

        # We can recover from this
        if not uei_region:
            self._custom_stats.passed_missing_uei_reads += 1
            return None

        if len(pid1_umi1_region) != self._pid_1_umi_1_region_len:
            self._custom_stats.failed_partial_upi1_umi1_reads += 1
            return AmpliconBuilderFailureReason.PARTIAL_PID1_UMI1

        if len(pid2_umi2_region) != self._pid_2_umi_2_region_len:
            self._custom_stats.failed_partial_upi2_umi2_reads += 1
            return AmpliconBuilderFailureReason.PARTIAL_PID2_UMI2

        # We can recover from this
        if len(uei_region) != self._uei_region_len:
            self._custom_stats.passed_partial_uei_reads += 1
            return None

        return None

    def handle(self, read1, read2, info1, info2):
        """Handle the reads and their modification info.

        This method is a placeholder for any additional processing that might be needed
        before building the amplicon.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")

    def __call__(
        self,
        read1: SequenceRecord | None,
        read2: SequenceRecord | None,
        info1: ModificationInfo | None,
        info2: ModificationInfo | None,
    ) -> Optional[SequenceRecord]:
        """Build an amplicon from paired-end reads (read1, read2), or a single read if only one is present.

        Return the processed read or None if the read has been
        "consumed" (filtered or written to an output file)
        and should thus not be passed on to subsequent steps.

        :param read1: the forward read (or None)
        :param read2: the reverse read (or None)
        :param info1: the modification info for the forward read, (ignored)
        :param info2: the modification info for the reverse read, (ignored)
        :return: the created amplicon as a new SequenceRecord or None if filtered out
        """
        if read1 is None and read2 is None:
            return None

        amplicon, error = self.handle(read1, read2, info1, info2)

        # Write failed reads to the "failed" writer
        if error is not None:
            self._custom_stats.failed_reads += 1
            if self._writer is not None:
                if read1 is not None:
                    read1.name += f" {error.value}"
                if read2 is not None:
                    read2.name += f" {error.value}"
                # Add an error message to the header
                self._writer.write(read1, read2)
            return None

        if read1 is not None:
            r1_regions = self._scan_forward_read(read1)
            if r1_regions.lbs1 is None:
                self._custom_stats.passed_missing_lbs1_anchor += 1

        # Combine the regions into a single amplicon sequence
        seq = self._region_combiner.build_sequence(
            amplicon.pid1_umi1_region_seq,
            amplicon.pid2_umi2_region_seq,
            amplicon.uei_region_seq,
        )
        qual = self._region_combiner.build_qualities(
            amplicon.pid1_umi1_region_qual,
            amplicon.pid2_umi2_region_qual,
            amplicon.uei_region_qual,
            amplicon.lbs1_region_qual,
            amplicon.lbs2_region_qual,
        )

        assert len(seq) == len(qual)

        # Create a new amplicon template to fill with the reads
        amplicon = SequenceRecord(
            name=read1.name if read1 is not None else read2.name,  # type: ignore
            sequence=seq.decode("ascii"),
            qualities=qual.decode("ascii"),
        )

        self._stats.update(amplicon.sequence)
        self._custom_stats.passed_reads += 1

        return amplicon


@dataclass
class Amplicon:
    """A dataclass to hold the regions of an amplicon."""

    pid1_umi1_region_seq: bytearray | Any | bytes | None
    pid1_umi1_region_qual: bytearray | Any | None
    pid2_umi2_region_seq: bytearray | Any | bytes | None
    pid2_umi2_region_qual: bytearray | Any | None
    uei_region_seq: bytearray | Any | bytes | None
    uei_region_qual: bytearray | Any | None
    lbs1_region_qual: bytearray | Any | None
    lbs2_region_qual: bytearray | Any | None


class PairedEndAmpliconBuilder(AmpliconBuilder):
    """A wrapper for AmpliconBuilder that handles paired-end reads."""

    def handle(
        self,
        read1: SequenceRecord,
        read2: SequenceRecord,
        info1: ModificationInfo | None = None,
        info2: ModificationInfo | None = None,
    ) -> tuple[Amplicon, AmpliconBuilderFailureReason | None]:
        """Process paired-end sequencing reads to build an Amplicon object and determine failure reasons.

        This method scans the provided forward and reverse reads for specific regions, generates consensus
        sequences and quality scores for each region, and checks for errors. If consensus cannot be reached,
        it returns an Amplicon object with all regions set to None and an appropriate failure reason.

        Args:
            read1 (SequenceRecord): The forward sequencing read.
            read2 (SequenceRecord): The reverse sequencing read.
            info1 (ModificationInfo | None, optional): Additional modification information for read1. Defaults to None.
            info2 (ModificationInfo | None, optional): Additional modification information for read2. Defaults to None.

        Returns:
            tuple[Amplicon, AmpliconBuilderFailureReason | None]:
                A tuple containing the constructed Amplicon object and a failure reason if applicable.
                If both reads are None, returns None.

        """
        r1_regions = self._scan_forward_read(read1)
        r2_regions = self._scan_reverse_read(read2)
        try:
            # Combine the info from forward and reverse reads, or use only the available read
            pid1_umi1_region_seq, pid1_umi1_region_qual = self._consensus_seq(
                read1,
                read2,
                r1_regions.pid1_umi1,  # type: ignore
                r2_regions.pid1_umi1,  # type: ignore
            )
            pid2_umi2_region_seq, pid2_umi2_region_qual = self._consensus_seq(
                read1,
                read2,
                r1_regions.pid2_umi2,  # type: ignore
                r2_regions.pid2_umi2,  # type: ignore
            )
            uei_region_seq, uei_region_qual = self._consensus_seq(
                read1,
                read2,
                r1_regions.uei,  # type: ignore
                r2_regions.uei,  # type: ignore
            )
            lbs1_region_qual = self._consensus_qual_lbs1(
                read1,
                read2,
                r1_regions.lbs1,  # type: ignore
                r2_regions.lbs1,  # type: ignore
            )
            lbs2_region_qual = self._consensus_qual_lbs2(
                read1,
                read2,
                r1_regions.lbs2,  # type: ignore
                r2_regions.lbs2,  # type: ignore
            )

            # Check for errors and increment the statistics
            error = self._check_regions(
                pid1_umi1_region_seq, uei_region_seq, pid2_umi2_region_seq
            )
        except AssertionError:
            error = AmpliconBuilderFailureReason.NO_CONSENSUS
            return Amplicon(
                pid1_umi1_region_seq=None,
                pid1_umi1_region_qual=None,
                pid2_umi2_region_seq=None,
                pid2_umi2_region_qual=None,
                uei_region_seq=None,
                uei_region_qual=None,
                lbs1_region_qual=None,
                lbs2_region_qual=None,
            ), error

        return Amplicon(
            pid1_umi1_region_seq=pid1_umi1_region_seq,
            pid1_umi1_region_qual=pid1_umi1_region_qual,
            pid2_umi2_region_seq=pid2_umi2_region_seq,
            pid2_umi2_region_qual=pid2_umi2_region_qual,
            uei_region_seq=uei_region_seq,
            uei_region_qual=uei_region_qual,
            lbs1_region_qual=lbs1_region_qual,
            lbs2_region_qual=lbs2_region_qual,
        ), error


class SingleEndAmpliconBuilder(AmpliconBuilder):
    """A wrapper for AmpliconBuilder that handles single-end reads."""

    def handle(
        self,
        read1: SequenceRecord | None,
        read2: SequenceRecord | None,
        info1: ModificationInfo | None = None,
        info2: ModificationInfo | None = None,
    ) -> tuple[Amplicon, AmpliconBuilderFailureReason | None]:
        """Process a single read or a pair of reads to construct an amplicon.

        Args:
            read1 (SequenceRecord | None): The forward read, or None if not available.
            read2 (SequenceRecord | None): The reverse read, or None if not available.
            info1 (ModificationInfo | None, optional): Modification info for the forward read (ignored).
            info2 (ModificationInfo | None, optional): Modification info for the reverse read (ignored).

        Returns:
            tuple[Amplicon, AmpliconBuilderFailureReason | None]:
                A tuple containing the constructed Amplicon and an optional failure reason.

        """
        is_read1 = read1 is not None
        read = read1 if is_read1 else read2

        regions = (
            self._scan_forward_read(read) if is_read1 else self._scan_reverse_read(read)  # type: ignore
        )

        is_reversed = not is_read1

        pid1_umi1_region_seq, pid1_umi1_region_qual = self._get_region_sequence(
            read,
            regions.pid1_umi1,  # type: ignore
            is_reversed=is_reversed,
        )
        pid2_umi2_region_seq, pid2_umi2_region_qual = self._get_region_sequence(
            read,
            regions.pid2_umi2,  # type: ignore
            is_reversed=is_reversed,
        )
        uei_region_seq, uei_region_qual = self._get_region_sequence(
            read,
            regions.uei,  # type: ignore
            is_reversed=is_reversed,
        )
        lbs1_region_qual = self._consensus_qual_lbs1(
            read1,
            read2,
            region1_slice=regions.lbs1 if is_read1 else None,  # type: ignore
            region2_slice=regions.lbs1 if not is_read1 else None,  # type: ignore
        )
        lbs2_region_qual = self._consensus_qual_lbs2(
            read1,
            read2,
            region1_slice=regions.lbs2 if is_read1 else None,  # type: ignore
            region2_slice=regions.lbs2 if not is_read1 else None,  # type: ignore
        )
        error = self._check_regions(
            pid1_umi1_region_seq,  # type: ignore
            uei_region_seq,  # type: ignore
            pid2_umi2_region_seq,  # type: ignore
        )

        return Amplicon(
            pid1_umi1_region_seq=pid1_umi1_region_seq,
            pid1_umi1_region_qual=pid1_umi1_region_qual,
            pid2_umi2_region_seq=pid2_umi2_region_seq,
            pid2_umi2_region_qual=pid2_umi2_region_qual,
            uei_region_seq=uei_region_seq,
            uei_region_qual=uei_region_qual,
            lbs1_region_qual=lbs1_region_qual,
            lbs2_region_qual=lbs2_region_qual,
        ), error



"""Model for report data returned by the single-cell analysis stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import pydantic

from pixelator.pna.report.models.base import SampleReport


class ProximityReport(pydantic.BaseModel):
    """Model for report data returned by the proximity analysis."""

    pass


class KCoreReport(pydantic.BaseModel):
    """Model for report data returned by the k-core analysis."""

    median_average_k_core: float


class SvdReport(pydantic.BaseModel):
    """Model for report data returned by the k-core analysis."""

    median_variance_explained_3d: float


class AnalysisSampleReport(SampleReport):
    """Model for report data returned by the single-cell analysis stage."""

    report_type: str = "analysis"

    proximity: ProximityReport | None
    k_cores: KCoreReport | None
    svd: SvdReport | None


class DenoiseSampleReport(SampleReport):
    """Model for report data returned by the single-cell denoise stage."""

    report_type: str = "denoise"
    number_of_umis_removed: int | None
    ratio_of_umis_removed: float | None
    number_of_disqualified_components: int | None
    ratio_of_disqualified_components: float | None



"""
Console script for pixelator

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from pathlib import Path

import click

from pixelator.common.utils import create_output_stage_dir, log_step_start, timer
from pixelator.mpx.cli.common import output_option
from pixelator.mpx.report.qcreport import make_report

logger = logging.getLogger(__name__)


@click.command(
    "report",
    short_help=("create a summary web report for all the samples"),
    options_metavar="<options>",
)
@click.argument(
    "input_folder",
    required=True,
    type=click.Path(exists=True),
    metavar="FOLDER",
)
@click.option(
    "--panel",
    required=False,
    type=str,
    help=(
        "A key of a panel file in the config,"
        " or a csv file with the antibody panel conjugations"
    ),
)
@click.option(
    "--metadata",
    required=False,
    default=None,
    type=click.Path(exists=True),
    help="Path to file with the samples metadata",
)
@output_option
@click.pass_context
@timer
def report(
    ctx,
    input_folder,
    panel,
    metadata,
    output,
):
    """
    Create a summary web report for all the samples (must complete all steps first)
    """
    # check that the input folder contains all the steps
    # (sub-folders preqc, adapterqc, demux, collapse, cluster and annotate)
    required_sub_folders = [
        "preqc",
        "adapterqc",
        "demux",
        "collapse",
        "graph",
        "annotate",
    ]
    if not all(
        map(
            lambda dir: (Path(input_folder) / dir).is_dir(),
            required_sub_folders,
        )
    ):
        raise click.ClickException(f"The structure of {input_folder} is not valid")

    # create output folder if it does not exist
    report_output = create_output_stage_dir(output, "report")

    # log input parameters
    log_step_start(
        "report",
        input_folder=input_folder,
        output=str(report_output),
        metadata=metadata,
    )

    # create html reports
    logger.info(f"Creating report for data present in {input_folder}")

    make_report(
        input_path=input_folder,
        output_path=str(report_output),
        panel=panel,
        metadata=metadata,
        verbose=ctx.obj["VERBOSE"],
    )



"""Workflow test helper for single-cell preqc command.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BasePreQCTestsMixin(BaseWorkflowTestMixin):
    """Base class for preqc command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_preqc_run"])
    ```
    """

    __stage_key__ = "preqc"

    @pytest.mark.dependency(name="test_preqc_run", depends=["test_amplicon_run"])
    def test_preqc_run(self):
        """Test and run the preqc command."""
        design = self.__get_data("design")
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")

        input_files = list(
            map(
                lambda f: str(f),
                (self.workdir / "amplicon").glob("*.merged.fastq.gz"),
            )
        )

        command = [
            "--log-file",
            "preqc-pixelator.log",
            "single-cell-mpx",
            "preqc",
            "--output",
            ".",
            "--design",
            design,
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        command += input_files

        self.context.run_command("preqc", command, input_files)
        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_preqc_run"])
    def test_preqc_logfile_exist(self):
        """Check if the log file exists."""
        assert (self.workdir / "preqc-pixelator.log").is_file()

    @pytest.mark.dependency(depends=["test_preqc_run"])
    def test_preqc_results_folder_exists(self):
        """Check if the output directory is created."""
        assert (self.workdir / "preqc").is_dir()

    @pytest.mark.dependency(depends=["test_preqc_run"])
    def test_preqc_processed_output_exists(self):
        """Check if the per antibody "processed" fastq files exist."""
        processed_files = (self.workdir / "preqc").glob("*.processed.fastq.gz")
        for f in processed_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_preqc_run"])
    def test_preqc_failed_output_exists(self):
        """Check if the "failed" fastq files exist."""
        failed_files = (self.workdir / "preqc").glob("*.failed.fastq.gz")
        for f in failed_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_preqc_run"])
    def test_preqc_html_output_exists(self):
        """Check if the fastp html report exist."""
        html_files = (self.workdir / "preqc").glob("*.report.html")
        for f in html_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_preqc_run"])
    def test_preqc_json_output_exists(self):
        """Check if the report.json summary exist."""
        json_files = (self.workdir / "preqc").glob("*.report.json")
        for f in json_files:
            assert f.is_file()



"""Custom pytest collector utilities to generate tests from yaml files.

Copyright © 2023 Pixelgen Technologies AB.
"""

import pytest

from .config import WorkflowConfig
from .workflow import PixelatorWorkflowTest


class NoModuleClass(pytest.Class):  # noqa
    """Mock pytest.Class that does not require a module as a parent."""

    class MockModule:
        """Fake module with a None obj."""

        def __init__(self):  # noqa
            self.obj = None

    def getparent(self, cls):  # noqa
        parent = super().getparent(cls)
        if parent is None:
            return self.MockModule()
        return parent


class YamlIntegrationTestsCollector(pytest.File):
    """Custom pytest collector generating tests from yaml files."""

    def collect(self):
        """Convert yaml files to test cases."""
        config = WorkflowConfig(self.path)

        for case in config.keys():
            name = case.split("-")[1]
            case_config = config.get_test_config(case)
            wf = type(
                f"TestWorkflow{name}",
                (PixelatorWorkflowTest,),
                {"test_id": case, **case_config},
            )

            collector = NoModuleClass.from_parent(parent=self, name=self.path.stem)
            # Need to add this explicitly since the constructor does not set obj
            collector.obj = wf

            for t in collector.collect():
                if t.parent is None:
                    t.parent = self
                t.fixturenames.append("use_workflow_context")
                yield t



"""Utility functions.

Copyright © 2024 Pixelgen Technologies AB.
"""

import re
import time
import typing
from functools import wraps
from pathlib import Path, PurePath
from typing import Iterable, Literal

import duckdb as dd
import pandas as pd
import polars as pl

from pixelator.common.utils import (
    R1_REGEX,
    R2_REGEX,
    get_extension,
    get_sample_name,
    logger,
)

_KNOWN_SUFFIXES = {".fastq", ".fq", ".fasta", ".fa"}
_KNOWN_COMPRESSION = {".gz", ".bz2", ".zst", ".xz"}

R1_REGEX = R"(.[Rr]1$)|(_[Rr]?1$)|(_[Rr]?1)(?P<suffix>_[0-9]{3})$"
R2_REGEX = R"(.[Rr]2$)|(_[Rr]?2$)|(_[Rr]?2)(?P<suffix>_[0-9]{3})$"


def _check_extensions(read: str | Path):
    allowed_extensions = {
        ".fastq.gz",
        ".fq.gz",
        ".fastq",
        ".fq",
        ".fastq.zst",
        ".fq.zst",
    }

    read = str(read)

    if not any(read.endswith(e) for e in allowed_extensions):
        raise ValueError(
            "Invalid file extension: expected .fq or .fastq (with .gz or .zst compression)"
        )


def get_read_sample_name(read: str | Path) -> str:
    """Extract the sample name from a read file.

    Strip fq.gz or fastq.gz extension and remove R1/R2 suffixes.
    Supported R1 R2 identifieds are:

    _R1,_R2 | _r1, _r2 | _1, _2 | .R1, .R2 | .r1, .r2

    :param read: filename of a fastq read file
    :return str: sample name
    :raise ValueError: if the read file does not have a valid extension
    """
    # group input file by sample id and order reads by R1 and R2
    _check_extensions(read)

    read_stem = Path(read).name
    read_stem = read_stem.removesuffix(get_extension(read_stem, 2)).rstrip(".")
    r1_match = re.search(R1_REGEX, read_stem)
    r2_match = re.search(R2_REGEX, read_stem)

    # Check if the r1 and r2 suffixes are "exclusive or"
    if r1_match and r2_match or (not r1_match and not r2_match):
        raise ValueError("Invalid R1/R2 suffix.")

    # We need to cast away the optional here r1 or r2 will always
    # return a match object since we checked for both being None above
    match = typing.cast(re.Match[str], r1_match or r2_match)

    # Remove the R1 or R2 suffix by using the indices returned by the match
    s, e = match.span()
    sample_name = read_stem[0:s] + read_stem[e:-1]

    if match.groupdict().get("suffix"):
        sample_name += match.group("suffix")

    return sample_name


def is_read_file(read: Path | str, read_type: Literal["r1"] | Literal["r2"]) -> bool:
    """Check if a read filename matches the specified read_type.

    Detects the presence of a common read 1 or read 2 suffix in the filename.

    :param read: filename of a fastq read file
    :param read_type: the read type to check for (r1 or r2)
    :return bool: True if the read file is a read 1 or 2 file
    :raise ValueError: if the read file does not have a valid extension
    :raise AssertionError: if the read_type is not 'r1' or 'r2'
    """
    read = Path(read).name

    if read_type not in ("r1", "r2"):
        raise AssertionError("Invalid read type: expected 'r1' or 'r2'")

    _check_extensions(read)

    match: re.Match[str] | None = None
    read_stem = Path(read.removesuffix(get_extension(read, 2)).rstrip(".")).name
    if read_type == "r1":
        match = re.search(R1_REGEX, read_stem)
    elif read_type == "r2":
        match = re.search(R2_REGEX, read_stem)
    else:
        raise AssertionError(
            "Invalid read type: could not find a read suffix in filename."
        )

    if not match:
        return False

    return True


def clean_suffixes(path: PurePath) -> PurePath:
    """Remove known suffixes and compression extensions from a path.

    :param path: The path to clean
    :return: The path without fasta/fastq suffixes or compression extensions
    """
    while path.suffix in _KNOWN_COMPRESSION:
        path = path.with_suffix("")

    while path.suffix in _KNOWN_SUFFIXES:
        path = path.with_suffix("")

    return path


def get_demux_filename_info(filename: str | Path | PurePath) -> tuple[str, int]:
    """Extract the sample name and part for a `demux` output parquet file.

    The demux output file are expeted to use following schema:
    <sample_name>.demux.part_<part_number>.parquet

    :param filename: path to the file
    :returns str: the demux part
    """
    sample_name = get_sample_name(filename)
    filename_str = str(filename)

    if (".demux" not in filename_str) or (".part_" not in filename_str):
        raise ValueError("Invalid demux filename. Did not contain .demux or .part_")

    match = re.match(r".*\.part_(\d+)\.*", filename_str)
    if match:
        demux_part = int(match.group(1))
        return sample_name, demux_part

    raise ValueError("Invalid demux filename.")


def timer(command_name: str | None = None):
    """Time the different steps of a function."""

    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwds):
            start_time = time.perf_counter()
            res = func(*args, **kwds)
            run_time = time.perf_counter() - start_time
            name = command_name or func.__name__
            logger.info("Finished pixelator %s in %.2fs", name, run_time)
            return res

        return inner

    return wrapper


def normalize_input_to_set(
    one_or_more_values: Iterable[str] | str | None,
) -> set[str] | None:
    """Normalize input to a set of strings."""
    if one_or_more_values is None:
        return None
    if isinstance(one_or_more_values, str):
        return {one_or_more_values}
    if isinstance(one_or_more_values, pd.Series):
        # For series return all truthy values from index
        return set(one_or_more_values[one_or_more_values].index)
    if isinstance(one_or_more_values, pl.Series):
        return set(one_or_more_values)
    if isinstance(one_or_more_values, pl.DataFrame):
        # if it is polars DataFrame with only one column get that
        if len(one_or_more_values.columns) == 1:
            return set(one_or_more_values.get_columns()[0])
        raise ValueError("If you pass a Polars DataFrame must have only one column")

    return {v for v in one_or_more_values}


def normalize_input_to_list(
    one_or_more_values: Iterable[str] | str | None,
) -> list[str] | None:
    """Normalize input to a list of strings."""
    if one_or_more_values is None:
        return None
    if isinstance(one_or_more_values, str):
        return [one_or_more_values]
    if isinstance(one_or_more_values, pd.Series):
        # For series return all truthy values from index
        return list(one_or_more_values[one_or_more_values].index)
    if isinstance(one_or_more_values, pl.Series):
        return one_or_more_values.to_list()
    if isinstance(one_or_more_values, pl.DataFrame):
        # if it is polars DataFrame with only one column get that
        if len(one_or_more_values.columns) == 1:
            return one_or_more_values.get_columns()[0].to_list()
        raise ValueError("If you pass a Polars DataFrame must have only one column")

    return [v for v in one_or_more_values]


def init_duckdb_conn(
    path: Path | str = ":memory:",
    read_only: bool = False,
    memory_limit: int | None = None,
    threads: int | None = None,
    temp_dir: str | Path | None = None,
    temp_dir_size_limit: str | None = None,
) -> dd.DuckDBPyConnection:
    """Initialize a duckdb connection with resource limits.

    Args:
        path: The path to the duckdb database file. Defaults to ":memory:" for in-memory database.
        read_only: Whether to open the database in read-only mode. Defaults to False.
        memory_limit: The memory limit in bytes. If None, no limit is set. Defaults to None.
        threads: The number of threads to use. If None, duckdb will decide. Defaults to None.
        temp_dir: The directory to use for temporary files. If None, duckdb will decide (defaults to /tmp). Defaults to None.
        temp_dir_size_limit: The maximum size of the temporary directory. If None, no limit is set. Defaults to None.

    Returns:
        A duckdb connection object.

    """
    conn = dd.connect(database=str(path), read_only=read_only)

    commands = []
    if memory_limit is not None:
        commands.append(f"SET memory_limit = '{memory_limit / 10**6}MiB';")
        logger.debug("Using DuckDB memory limit: %s MB", memory_limit / 10**6)
    if threads is not None:
        commands.append(f"SET threads = {threads};")
        logger.debug("Using DuckDB threads limit: %s", threads)
    if temp_dir is not None:
        temp_dir = str(Path(temp_dir).absolute())
        commands.append(f"SET temp_directory = '{temp_dir}';")
        logger.debug("Using DuckDB temp directory: %s", temp_dir)
    if temp_dir_size_limit is not None:
        commands.append(f"SET max_temp_directory_size = '{temp_dir_size_limit}';")
        logger.debug("Using DuckDB temp directory size limit: %s", temp_dir_size_limit)

    if commands:
        conn.execute("\n".join(commands))

    conn = dd.connect(":memory:")
    return conn



"""Copyright © 2023 Pixelgen Technologies AB."""

from pixelator.common.report.models import SummaryStatistics

from .base import SampleReport
from .collapse import CollapseSampleReport
from .commands import CommandInfo, CommandOption
from .reads_flow import ReadsDataflowReport

__all__ = [
    "SampleReport",
    "CollapseSampleReport",
    "ReadsDataflowReport",
    "CommandInfo",
    "CommandOption",
    "SummaryStatistics",
]



"""Model for flow of input/output of molecule counts through processing stages.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import textwrap

import pydantic

from pixelator.mpx.report.models.base import SampleReport


class MoleculesDataflowReport(SampleReport):
    """Model for flow of input/output counts through processing stages."""

    raw_molecule_count: int = pydantic.Field(
        ...,
        description=textwrap.dedent(
            """The number of raw unique molecules in the sample.
            This is the number of unique molecules after deduplication of reads with
            close UPIA UPIA and UMI sequences by the collapse stage.
            """
        ),
    )

    size_filter_fail_molecule_count: int = pydantic.Field(
        ...,
        description=(
            "The number of molecules in components that do NOT pass "
            "the component size filters."
        ),
    )

    aggregate_molecule_count: int | None = pydantic.Field(
        ...,
        description="The number of molecules in components identified as aggregates.",
    )

    cell_molecule_count: int = pydantic.Field(
        ...,
        description=textwrap.dedent(
            """
            The number of molecules in the graph after annotate.
            Note that this includes the molecules in aggregate components.
            """
        ),
    )

    @pydantic.computed_field(  # type: ignore
        return_type=float,
        description="Return the fraction of raw input reads in unique molecules.",
    )
    @property
    def fraction_molecules_in_cells(self) -> float:
        """Return the fraction of raw input reads in unique molecules."""
        return self.cell_molecule_count / self.raw_molecule_count

    @pydantic.computed_field(  # type: ignore
        return_type=float,
        description=textwrap.dedent(
            """Return the fraction of raw molecules that were discarded.
            This is equal to: 1 - `fraction_molecules_in_cells`.
            """
        ),
    )
    @property
    def fraction_molecules_discarded(self) -> float:
        """Return the fraction of raw molecules that were discarded.

        This is 1 - `fraction_molecules_in_cells`.
        """
        return 1 - self.cell_molecule_count / self.raw_molecule_count



"""Console script for pixelator (demux).

Copyright © 2022 Pixelgen Technologies AB.
"""

from pathlib import Path

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.common.utils.units import parse_size
from pixelator.pna.cli.common import (
    design_option,
    logger,
    memory_option,
    output_option,
    panel_option,
    threads_option,
)
from pixelator.pna.config import pna_config
from pixelator.pna.config.panel import load_antibody_panel
from pixelator.pna.demux import (
    correct_marker_barcodes,
    demux_barcode_groups,
    finalize_batched_groups,
)
from pixelator.pna.demux.report import DemuxSampleReport


def _chunk_size_validator(ctx, param, value):
    try:
        return int(parse_size(value))
    except ValueError as exc:
        raise click.BadParameter(
            "chunk size must be a positive integer, optionally with a unit suffix [K, M, G]"
        )


@click.command(
    "demux",
    short_help="demultiplex pixel data (FASTQ) to generate one file per antibody",
    options_metavar="<options>",
)
@click.argument(
    "fastq_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="FASTQ",
)
@click.option(
    "--mismatches",
    default=1,
    required=False,
    show_default=True,
    help="The number of mismatches allowed in marker barcodes",
)
@click.option(
    "--output-chunk-reads",
    default="10M",
    type=click.STRING,
    callback=_chunk_size_validator,
    required=False,
    show_default=True,
    help="The target number of molecules in each output parquet file",
)
@click.option(
    "--output-max-chunks",
    default=8,
    type=click.INT,
    required=False,
    show_default=True,
    help="The maximum number of marker parts to split the demuxed data into",
)
@click.option(
    "--strategy",
    type=click.Choice(["paired", "independent"]),
    default="independent",
    required=False,
    help=(
        "The strategy for splitting demuxed files. 'paired' will split files by batches of PID pairs"
        "'independent' will create two partitions of files split by marker1 and marker2 independently."
    ),
)
@threads_option
@memory_option
@design_option
@panel_option
@output_option
@click.pass_context
@timer
def demux(
    ctx,
    fastq_file,
    mismatches,
    output_chunk_reads,
    output_max_chunks,
    panel,
    output,
    design,
    threads,
    memory,
    strategy,
):
    """Demultiplex Molecular Pixelation data (FASTQ) to generate one file per antibody."""
    # log input parameters
    input_files = [fastq_file]
    log_step_start(
        "demux",
        input_files=input_files,
        output=output,
        mismatches=mismatches,
        output_chunk_reads=output_chunk_reads,
        output_max_chunks=output_max_chunks,
        panel=panel,
        design=design,
        strategy=strategy,
    )

    # Respect the global --cores if provided
    # TODO: Remove threads after unifying all CLI commands to use --cores
    threads = threads if threads > 0 else ctx.obj["CORES"]
    duckdb_tmp_dir = ctx.obj.get("DUCKDB_TEMP_DIR")

    # some basic sanity check on the input files
    sanity_check_inputs(
        input_files,
        allowed_extensions=("fastq.gz", "fq.gz", "fastq", "fq", "fastq.zst", "fq.zst"),
    )

    fastq_file = Path(fastq_file)

    # create output folder if it does not exist
    demux_output = create_output_stage_dir(output, "demux")

    # load assay design
    assay = pna_config.get_assay(design)
    # load marker panel
    panel = load_antibody_panel(pna_config, panel)

    logger.info(f"Correcting marker barcodes for input: {fastq_file}")

    stats, corrected, failed = correct_marker_barcodes(
        input=input_files[0],
        assay=assay,
        panel=panel,
        output=demux_output,
        save_failed=True,
        threads=threads,
    )

    # Store intermediate parquet files before deduplication and sorting
    tmp_output_dir = demux_output / "tmp"
    tmp_output_dir.mkdir()

    demux_barcode_groups(
        corrected_reads=corrected,
        assay=assay,
        panel=panel,
        stats=stats,
        output_dir=tmp_output_dir,
        threads=threads,
        reads_per_chunk=output_chunk_reads,
        max_chunks=output_max_chunks,
        stategy=strategy,
    )

    finalize_batched_groups(
        input_dir=tmp_output_dir,
        output_dir=demux_output,
        strategy=strategy,
        memory=memory,
        threads=threads,
        temp_dir=duckdb_tmp_dir,
    )

    # remove results in the temporary output directory
    for file in tmp_output_dir.iterdir():
        file.unlink()
    tmp_output_dir.rmdir()

    sample_name = get_sample_name(fastq_file)
    report_json = demux_output / f"{sample_name}.report.json"

    write_parameters_file(
        ctx,
        demux_output / f"{sample_name}.meta.json",
        command_path="pixelator single-cell-pna demux",
    )

    report = DemuxSampleReport(
        product_id="single-cell-pna",
        report_type="demux",
        sample_id=sample_name,
        **stats.as_json(),
    )
    report.write_json_file(report_json, indent=4)
